from collections import OrderedDict

from glypy.structure import Glycan, Monosaccharide
from glypy.structure.glycan_composition import GlycanComposition
from glypy.utils import tree

from .node_type import NodeTypeSpec
from .utils import base52


class WURCSWriter(object):
    """Implementation of WURCS encoding process.

    Includes the steps for creating each section of the WURCS encoding,
    and populating them from a saccharide structure, composition, or
    monosaccharide.

    """

    version = '2.0'

    def __init__(self, glycan):
        self.glycan = glycan
        self.node_type_map = OrderedDict()
        self.node_index_to_node_type = OrderedDict()
        self.index_to_glyph = dict()
        self.id_to_index = dict()
        self.extract_node_types()

    def extract_node_types(self):
        node_types = OrderedDict()
        node_index_to_node_type = OrderedDict()
        index_to_glyph = dict()
        id_to_index = dict()
        for i, node in enumerate(self._iter_monosaccharides(), 1):
            node_type = NodeTypeSpec.from_monosaccharide(node)
            index_to_glyph[i] = base52(i - 1)
            id_to_index[node.id] = i
            node_index_to_node_type[i] = node_type
            if node_type not in node_types:
                node_types[node_type] = len(node_types) + 1
        self.node_type_map = node_types
        self.node_index_to_node_type = node_index_to_node_type
        self.index_to_glyph = index_to_glyph
        self.id_to_index = id_to_index

    def format_version(self):
        return "WURCS=%s" % (self.version, )

    def _iter_monosaccharides(self):
        if isinstance(self.glycan, GlycanComposition):
            for key, value in self.glycan.items():
                for i in range(value):
                    yield key
        else:
            for x in self.glycan.iternodes():
                yield x

    def _iter_links(self):
        if isinstance(self.glycan, GlycanComposition):
            pass
        else:
            for x in self.glycan.iterlinks():
                yield x

    def format_count_section(self):
        count_nodes = len(list(self._iter_monosaccharides()))
        count_links = len(list(self._iter_links()))
        count_section = "%s,%s,%s" % (len(self.node_type_map), count_nodes, count_links)
        if isinstance(self.glycan, GlycanComposition):
            count_section += '+'
        return count_section

    def format_node_types(self):
        return ''.join('[%s]' % (str(s),) for s in self.node_type_map.keys())

    def format_node_type_index(self):
        node_type_sequence = []
        for index, node_type in self.node_index_to_node_type.items():
            node_type_sequence.append(self.node_type_map[node_type])
        return '-'.join(map(str, node_type_sequence))

    def format_links(self):
        links = []
        if isinstance(self.glycan, GlycanComposition):
            return ""
        for _, link in self.glycan.iterlinks():
            parent_index = self.id_to_index[link.parent.id]
            child_index = self.id_to_index[link.child.id]
            parent_glyph = self.index_to_glyph[parent_index]
            child_glyph = self.index_to_glyph[child_index]
            parent_position = link.parent_position
            if parent_position == -1:
                parent_position = '?'
            child_position = link.child_position
            if child_position == -1:
                child_position = '?'
            link_spec = '%s%s-%s%s' % (parent_glyph, parent_position, child_glyph, child_position)
            links.append(link_spec)
        return '_'.join(links)

    def write(self):
        self.extract_node_types()
        sections = [self.format_version(), self.format_count_section(), self.format_node_types(),
                    self.format_node_type_index(), ]
        if not isinstance(self.glycan, GlycanComposition):
            sections.append(self.format_links())
        return '/'.join(sections)


def dumps(glycan):
    """Encode a saccharide object as a WURCS 2.0 string.

    .. note::
        The WURCS canonicalization has not been implemented yet, so the generated
        string may differ from other sources. However, the :mod:`~.canonicalize` module
        can be used to standardize structures prior to encoding them.

    Parameters
    ----------
    glycan : :class:`~.Glycan`, :class:`~.GlycanComposition`, or :class:`~.Monosaccharide`
        The structure to encode

    Returns
    -------
    :class:`str`
        The structure encoded as a string.
    """
    if not isinstance(glycan, GlycanComposition):
        try:
            glycan = tree(glycan)
        except TypeError:
            if isinstance(glycan, Monosaccharide):
                nts = NodeTypeSpec.from_monosaccharide(glycan)
                return nts.to_res()
            else:
                raise
    return WURCSWriter(glycan).write()


Glycan.register_serializer('wurcs', dumps)
Monosaccharide.register_serializer('wurcs', dumps)



'''A set of algorithms for finding similar structures and substructures.
'''

from .inclusion import (
    topological_inclusion, exact_ordering_inclusion, subtree_of,
    find_matching_subtree_roots, walk_with, TopologicalInclusionMatcher)

from .common_subgraph import (
    maximum_common_subgraph, treelets, n_saccharide_similarity, distinct_fragments,
    treelet_enrichment)

from .common_subgraph import (
    Treelet, TreeletIterator, TreeletEnrichmentTest,
    MaximumCommonSubgraphSolver, MaximumCommonSubtreeResults)



from .ec import (
    EnzymeInformation, EnzymeCommissionNumber, EnzymeDatabase, expasy_enzyme_db)

from .graph import (
    EnzymeEdge, EnzymeGraph, GlycanCompositionEnzymeGraph,
    GlycanStructureEnzymeGraph, _enzyme_graph_inner)

from .pathways import (
    Glycoenzyme, Glycosylase, Glycosyltransferase,
    Substituentransferase, rejecting, reject_on_path,
    make_n_glycan_pathway, make_mucin_type_o_glycan_pathway)

from .glycome import (Glycome, MultiprocessingGlycome)


__all__ = [
    "EnzymeDatabase", "EnzymeInformation", "EnzymeCommissionNumber",
    "Glycoenzyme", "Glycosylase", "Glycosyltransferase",
    "Substituentransferase", "rejecting", "reject_on_path",
    "make_n_glycan_pathway", "make_mucin_type_o_glycan_pathway",
    "Glycome", "MultiprocessingGlycome", "EnzymeGraph", "EnzymeEdge",
    "GlycanStructureEnzymeGraph", "GlycanCompositionEnzymeGraph",
    "_enzyme_graph_inner", "expasy_enzyme_db",
]



from collections import defaultdict
import operator

from six import string_types as basestring

from ...utils import groupby
from ...structure import (
    named_structures, Monosaccharide, Substituent,
    Anomer, Stem, RingType,
    SuperClass, Configuration)
from ...algorithms.similarity import (monosaccharide_similarity, has_substituent,
                                      has_modification, has_monosaccharide,
                                      is_generic_monosaccharide)
from ...composition.composition_transform import strip_derivatization
from .synonyms import monosaccharides as monosaccharide_synonyms


def has_ambiguity(node):
    ambiguous = node.stem[0] is Stem.x or node.anomer is Anomer.x or\
        node.superclass is SuperClass.x or node.ring_type is RingType.x
    return ambiguous


# A static copy of monosaccharide names to structures for copy-free comparison
monosaccharides = dict(named_structures.monosaccharides)
monosaccharides_ordered = sorted(list(monosaccharides.items()), key=lambda x: has_ambiguity(x[1]))


def get_preferred_name(name, selector=min, key=len):
    '''
    Given a name, of its synonyms, find the name that satisfies the `selector`
    criterion function (:func:`min`) based on some `key` function of the name (:func:`len`)

    Parameters
    ----------
    name: str
        Given name to compare to synonyms
    selector: function
        Function to use to select the preferred name by some statistic
    key: function
        Function to use to convert names into statistics

    Returns
    -------
    str
    '''
    preferred_name = selector(monosaccharide_synonyms.get(name, [name]) + [name], key=key)
    return preferred_name


def is_a(node, target, tolerance=0, include_modifications=True, include_substituents=True, exact=True,
         short_circuit=False, ignore_ring=True, **kwargs):
    '''
    Perform a semi-fuzzy match between `node` and `target` where node is the unqualified
    residue queried and target is the known residue to be matched against.

    Forwards all unmatched arguments to :func:`~.monosaccharide_similarity`

    Parameters
    ----------
    node: Monosaccharide or Substituent
        Object to be identified
    target: Monosaccharide, Substituent or str
        The reference type. May be a |str| object which is used to look up a |Monosaccharide| by name in
        :obj:`glypy.monosaccharides`
    tolerance: int
        The error tolerance for the search
    include_modifications: bool
        Whether or not to include modifications in comparison. Defaults to |True|
    include_substituents: bool
        Whether or not to include substituents in comparison. Defaults to |True|
    exact: bool
        Whether or not to penalize for unmatched attachments. Defaults to |True|
    Returns
    -------
    bool

    '''
    res = 0
    qs = 0
    if isinstance(target, basestring):
        target = monosaccharides[target]

    if isinstance(node, Substituent):
        if not isinstance(target, Substituent):
            return False
        else:
            res += node.name == target.name
            qs += 1
    else:
        if not isinstance(target, Monosaccharide):
            return False
        res, qs = monosaccharide_similarity(node, target, include_modifications=include_modifications,
                                            include_substituents=include_substituents,
                                            include_children=False, exact=exact, ignore_ring=ignore_ring,
                                            short_circuit_after=tolerance if short_circuit else None, **kwargs)
    threshold = (qs - res) <= tolerance
    return threshold


def identify(node, blacklist=None, tolerance=0, include_modifications=True, include_substituents=True,
             ignore_ring=True, **kwargs):
    '''
    Attempt to find a common usage name for the given |Monosaccharide|, `node`. The name is determined by
    performing an incremental comparison of the traits of `node` with each named residue in the database
    accessed at :obj:`glypy.monosaccharides`.

    Forwards all unmatched arguments to :func:`~.monosaccharide_similarity`

    Parameters
    ----------
    node: Monosaccharide
        Object to be identified
    blacklist: list
        The set of all monosaccharides to not attempt matching against, because they are too general.
    tolerance: int
        The error tolerance for the search
    include_modifications: bool
        Whether or not to include modifications in comparison. Defaults to |True|
    include_substituents: bool
        Whether or not to include substituents in comparison. Defaults to |True|

    Returns
    -------
    str

    Raises
    ------
    IdentifyException:
        When a suitable name cannot be found.

    See Also
    --------
    is_a
    preferred_name
    monosaccharide_similarity
    '''
    if blacklist is None:
        blacklist = {"Pen", "Hex", "Hep", "Oct", "Non"}
    for name, structure in monosaccharides_ordered:
        if name in blacklist:
            continue
        if is_a(node, structure, tolerance, include_modifications, include_substituents, ignore_ring=ignore_ring,
                **kwargs):
            return get_preferred_name(name)
    raise IdentifyException("Could not identify {}".format(node))


class IdentifyException(KeyError):
    pass


def naive_name_monosaccharide(monosaccharide):
    '''
    Generate a generic name for `monosaccharide`, based loosely on IUPAC
    naming schema without including information about linkage.

    The tendency for monosaccharides of superclass > 7 to have special names,
    which will be used preferentially if possible.

    Parameters
    ----------
    monosaccharide: Monosaccharide

    Returns
    -------
    str:
        A simple name based on `SuperClass`, modifications, and substituents.

    See Also
    --------
    :func:`glypy.io.nomenclature.identity.identify`

    '''
    try:
        c = monosaccharide.clone()
        if not isinstance(c, Monosaccharide):
            return None
        strip_derivatization(c)
        try:
            if monosaccharide.superclass.value > 6:
                return identify(c, tolerance=0)
        except Exception:
            pass
        c.anomer = None
        return identify(c)
    except IdentifyException:
        try:
            c.stem = None
            c.configuration = None
            return identify(c)
        except IdentifyException:
            return "".join(mod.name for mod in list(c.modifications.values()) if mod.name != 'aldi') +\
                   c.superclass.name.title() + ''.join([''.join(map(str.title, subst.name.split("_")))[:3]
                                                        for p, subst in c.substituents()])


def split_along_axis(monosaccharides, axis):
    getter = operator.attrgetter(axis)
    groups = groupby(monosaccharides, getter)
    return groups


def residue_list_to_tree(monosaccharides, axes=('anomer', 'superclass', 'stem', 'configuration')):
    root = split_along_axis(monosaccharides, axes[0])
    if len(axes) > 1:
        for level, group in list(root.items()):
            root[level] = residue_list_to_tree(group, axes[1:])
    return root


class MonosaccharideIdentifier(object):
    def __init__(self, reference_index=None, **kwargs):
        if reference_index is None:
            reference_index = dict(named_structures.monosaccharides)
        self.reference_index = dict(reference_index)
        self.trait_tree = residue_list_to_tree(set(self.reference_index.values()))
        self.name_map = self._build_name_map()

    def _build_name_map(self):
        by_monosaccharide = groupby(self.reference_index.items(), lambda x: x[1])
        monosaccharide_to_name = {
            k: min([vi[0] for vi in v], key=len)
            for k, v in by_monosaccharide.items()
        }
        return monosaccharide_to_name

    def _find_potential_matches(self, monosaccharide, exact_candidates=False, **kwargs):
        anomer = monosaccharide.anomer
        candidates = []
        members = self.trait_tree[anomer]
        candidates.append(members)
        if anomer != Anomer.x:
            candidates.append(self.trait_tree[Anomer.x])
        superclass = monosaccharide.superclass
        next_candidates = []
        for candidate in candidates:
            if not candidate:
                continue
            next_candidates.append(candidate[superclass])
            if superclass != SuperClass.x:
                next_candidates.append(candidate[SuperClass.x])
        candidates = next_candidates
        next_candidates = []
        stem = monosaccharide.stem
        for candidate in candidates:
            if not candidate:
                continue
            next_candidates.append(candidate[stem])
            if stem != (Stem.x, ):
                next_candidates.append(candidate[(Stem.x, )])
        candidates = next_candidates
        next_candidates = []
        configuration = monosaccharide.configuration
        for candidate in candidates:
            if not candidate:
                continue
            next_candidates.append(candidate[configuration])
            if configuration != (Configuration.x, ):
                next_candidates.append(candidate[(Configuration.x, )])
        candidates = []
        for c in next_candidates:
            candidates.extend(c)
        is_a_potential = {}
        kwargs.setdefault('exact', True)
        kwargs.setdefault('treat_null_as_wild', False)
        kwargs.setdefault('match_attachement_positions', True)
        for c in candidates:
            if is_a(monosaccharide, c, exact=exact_candidates):
                a, b = monosaccharide_similarity(
                    monosaccharide, c, **kwargs)
                is_a_potential[c] = a / float(b)
        return is_a_potential

    def query(self, monosaccharide, **kwargs):
        is_a_potential = self._find_potential_matches(monosaccharide, **kwargs)
        if not is_a_potential:
            return None
        match = max(is_a_potential.items(), key=lambda x: x[1])[0]
        return match

    def identify(self, monosaccharide, **kwargs):
        template = self.query(monosaccharide, **kwargs)
        if template is not None:
            return self.name_map[template]
        else:
            raise IdentifyException(monosaccharide)



import re

from glypy.structure import glycan_composition
from glypy.structure.glycan_composition import (
    FrozenGlycanComposition,
    FrozenMonosaccharideResidue,
    SubstituentResidue)
from glypy.structure.glycan import Glycan
from glypy.utils import invert_dict


try:
    import requests
except ImportError:
    requests = None


defined_symbols = {
    "Hex": FrozenMonosaccharideResidue.from_iupac_lite("Hex"),
    "HexNAc": FrozenMonosaccharideResidue.from_iupac_lite('HexNAc'),
    "dHex": FrozenMonosaccharideResidue.from_iupac_lite('dHex'),
    "NeuAc": FrozenMonosaccharideResidue.from_iupac_lite("NeuAc"),
    "NeuGc": FrozenMonosaccharideResidue.from_iupac_lite("NeuGc"),
    "S": SubstituentResidue("sulfate"),
    "P": SubstituentResidue("phosphate"),
    "Xyl": FrozenMonosaccharideResidue.from_iupac_lite("Xyl"),
    "HexA": FrozenMonosaccharideResidue.from_iupac_lite("HexA"),
}


monosaccharide_to_symbol = invert_dict(defined_symbols)


tokenizer = re.compile(r"([^:\s]+):(\d+)")


def loads(string):
    '''Parse a GlyConnect glycan composition into a :class:`~.FrozenGlycanComposition`

    Parameters
    ----------
    string: str
        The string to parse

    Returns
    -------
    :class:`~.FrozenGlycanComposition`

    Raises
    ------
    :class:`KeyError`: Raised if a key isn't defined by the GlyConnect dialect
    '''
    tokens = tokenizer.findall(string)
    gc = FrozenGlycanComposition()
    for mono, count in tokens:
        mono = defined_symbols[mono]
        count = int(count)
        gc[mono] += count
    return gc


def dumps(composition):
    '''Encode :class:`~.GlycanComposition` or :class:`~.Glycan` into the GlyConnect
    glycan composition text format.

    Parameters
    ----------
    composition: :class:`~.GlycanComposition` or :class:`~.Glycan`
        The structure to format

    Returns
    -------
    :class:`str`

    Raises
    ------
    :class:`KeyError`: Raised if a key isn't defined by the GlyConnect Compozitor dialect
    '''
    if isinstance(composition, Glycan):
        composition = FrozenGlycanComposition.from_glycan(composition)
    tokens = []
    for key, value in composition.items():
        key = monosaccharide_to_symbol[key]
        tokens.append("%s:%d" % (key, value))
    return ' '.join(tokens)


API_SERVER = "https://glyconnect.expasy.org/api"


def from_glytoucan_id(glytoucan_id):
    response = requests.post(
        "{server}/structures/search/glytoucan".format(server=API_SERVER),
        {"glytoucan_id": glytoucan_id})
    response.raise_for_status()
    data = response.json()
    return data




import re
import operator
from collections import defaultdict
from uuid import uuid4
from itertools import chain
import matplotlib.pyplot as plt
import matplotlib.transforms as mtransforms

from six import string_types as basestring

import numpy as np
from glypy import monosaccharides
from glypy.structure import constants
from glypy.utils import root
from glypy.io.nomenclature import identity
from glypy.algorithms.similarity import is_derivatized
from glypy.composition.composition_transform import strip_derivatization

from .buchheim import BalancedTreeLayout
from .topological_layout import TopologicalTreeLayout
from . import cfg_symbols, iupac_symbols, snfg_symbols
from .geometry import centroid, interpolate
from .fragment_annotation import BondCleavageArtist


line_to = cfg_symbols.CFGNomenclature().line_to


nomenclature_map = {
    "cfg": cfg_symbols.CFGNomenclature(),
    'iupac': iupac_symbols.IUPACTextSymbolicNomenclature(),
    "snfg": snfg_symbols.SNFGNomenclature()
}

layout_map = {
    "balanced": BalancedTreeLayout,
    "topological": TopologicalTreeLayout
}

DEFAULT_SYMBOL_SCALE_FACTOR = 0.17


#: :data:`special_cases` contains a list of names for
#: special case monosaccharides
_special_case_fucose = monosaccharides.Fuc
_special_case_fucose.configuration = None
_special_case_fucose.anomer = None
_special_case_xylose = monosaccharides.Xyl
_special_case_xylose.configuration = None
_special_case_xylose.anomer = None

special_cases = [_special_case_fucose, _special_case_xylose]


anomer_symbol_map = {
    constants.Anomer.alpha: r'\alpha',
    constants.Anomer.beta: r'\beta',
    constants.Anomer.uncyclized: r'o',
    constants.Anomer.x: r'?'
}


def sign(x):
    if x > 0:
        return 1
    else:
        return -1


def flatten(iterable):
    acc = []
    for item in iterable:
        try:
            acc.extend(flatten(item))
        except Exception:
            acc.append(item)
    return acc


def is_special_case(node):
    '''
    Check to see if `node` is a special case which requires a different layout
    scheme. See `special_cases`
    '''
    if is_derivatized(node):
        node = strip_derivatization(node.clone())
    for case in special_cases:
        if identity.is_a(node, case) and len(list(node.children())) == 0:
            return True
    return False


def box(node, ax=None):  # pragma: no cover
    if ax is None:
        ax = node.axes
    xmin, xmax, ymin, ymax = node.extrema()
    ax.plot((xmin - 0.2, xmax + 0.2), (ymin - 0.2, ymin - 0.2), c='b')
    ax.plot((xmin - 0.2, xmax + 0.2), (ymax + 0.2, ymax + 0.2), c='b')
    ax.plot((xmin - 0.2, xmin - 0.2), (ymin - 0.2, ymax + 0.2), c='b')
    ax.plot((xmax + 0.2, xmax + 0.2), (ymin - 0.2, ymax + 0.2), c='b')


def make_gid(s):  # pragma: no cover
    '''Formats a string to use ``kebab-case`` and sanitize illegal characters'''
    s = str(s)
    return re.sub(r'[\s\|\(\):]', '-', s)


def breadth_first_traversal(tree, visited=None):
    if visited is None:
        visited = set()
    if tree.id not in visited:
        visited.add(tree.id)
        yield tree
        for child in tree:
            for descend in breadth_first_traversal(child, visited=visited):
                yield descend


def _bounding_box_to_limits(dtree):
    xlim, ylim = dtree._compute_bounding_box()
    xlim = list(xlim)
    xlim[0] -= 2
    xlim[1] += 2
    ylim = list(ylim)
    ylim[0] -= 2
    ylim[1] += 2
    return np.array(xlim), np.array(ylim)


def enumerate_tree(tree, ax, labeler=None):
    """Label each node of `tree` on `ax` with the node's :attr:`id`

    Parameters
    ----------
    tree: :class:`DrawTreeNode`
        A drawn :class:`DrawTreeNode` object
    ax: matplotlib.Axes
        The axes on which `tree` has been drawn

    Returns
    -------
    tree, ax
    """
    if labeler is None:
        labeler = operator.attrgetter('id')
    nodes = list(breadth_first_traversal(tree))
    for node in nodes:
        x, y = node.coords()
        ax.text(x, y + 0.2, labeler(node.tree), color='red')
    # Update the plot
    ax.get_figure().canvas.draw()
    return tree, ax


def find_link(parent, child):  # pragma: no cover
    for p, link in parent.links.items():
        if child.parent_linkage == p:
            if link.child == child:
                return link


def get(tree, node_id):
    for node in breadth_first_traversal(tree):
        if node.tree.id == node_id:
            return node


def get_link_pair(tree, link_id):
    link = None
    for node in breadth_first_traversal(tree):
        for p, t_link in node.tree.links.items():
            if t_link.id == link_id:
                link = t_link
                break
    if link is None:
        raise Exception("Link not found")
    parent = get(tree, link.parent.id)
    child = get(tree, link.child.id)
    return parent, child


INF = float("inf")


class DrawTreeNode(object):
    def __init__(self, tree, parent=None, depth=0, number=1, parent_linkage=None, visited=None):
        if visited is None:
            visited = dict()
        self.id = tree.id
        if self.id in visited:
            raise Exception("Cycle detected")
        visited[self.id] = self
        self.x = -1.
        self.y = depth
        self.depth = depth
        self.tree = tree
        self.parent = parent
        self.parent_linkage = parent_linkage

        self._axes = None

        self.mask_special_cases = True
        # Unused. Artefact of Buccheim
        self.thread = None
        self.mod = 0
        self.ancestor = self
        self.change = self.shift = 0
        self._lmost_sibling = None

        # Number of the node in its group of siblings 1..n
        self.number = number

        if self.parent is not None:
            self.data = self.parent.data
            self.uuid = self.parent.uuid
        else:
            self.data = defaultdict(lambda: defaultdict(dict))
            self.uuid = uuid4().hex

        self.children = [self.resolve_creation_cycle(c[1], depth + 1, i + 1, c[0], visited)
                         for i, c in enumerate(sorted(tree.children(), key=lambda x: x[0]))]

        # A node owns all lines originating from it.
        self.lines = []

    def resolve_creation_cycle(self, node, depth, i, parent_linkage, visited):
        if node.id in visited:
            return visited[node.id]
        else:
            return self.__class__(node, self, depth + 1, i + 1, parent_linkage, visited)

    def __iter__(self):
        return iter(self.children)

    def traverse(self):
        return breadth_first_traversal(self)

    @property
    def children(self):
        if self.mask_special_cases:
            return [child for child in self._children if not (
                is_special_case(child.tree) and len(child.children) == 0)]
        else:
            return self._children

    @children.setter
    def children(self, value):
        self._children = value

    @property
    def axes(self):
        return self._axes

    @axes.setter
    def axes(self, value):
        self._axes = value
        for child in self:
            if child.axes != value:
                child.axes = value

    def fix_special_cases(self, offset=0.5, visited=None):
        if visited is None:
            visited = set()
        if self.id in visited:
            return
        visited.add(self.id)
        self.mask_special_cases = False
        for child in self.children:
            if is_special_case(child.tree):
                if child.parent_linkage > 3 or child.parent_linkage == -1:
                    child.y = self.y
                    child.x = self.x + offset
                elif child.parent_linkage > 1:
                    child.y = self.y
                    child.x = self.x - offset
                else:
                    raise Exception(
                        "Don't know how to handle special case child node %r, %r." % (
                            child, child.parent_linkage))
            child.fix_special_cases(offset, visited)

    def draw(self, at=(0, 0), ax=None, symbol_nomenclature="snfg", label=True, **kwargs):
        if isinstance(symbol_nomenclature, basestring):
            symbol_nomenclature = nomenclature_map[symbol_nomenclature]
        elif symbol_nomenclature is None:
            symbol_nomenclature = snfg_symbols.SNFGNomenclature()
        if ax is None:  # pragma: no cover
            ax = self.axes
            if ax is None:
                fig, ax = plt.subplots()
        self.draw_branches(at=at, ax=ax, symbol_nomenclature=symbol_nomenclature,
                           label=label, visited=set(), **kwargs)
        self.draw_nodes(at=at, ax=ax, symbol_nomenclature=symbol_nomenclature,
                        label=label, visited=set(), **kwargs)

    def coords(self, at=None):
        if at is None:
            at = (0, 0)
        if self.transform:
            at = self.transform.transform(at)
            x, y = self.transform.transform((self.x, self.y))
        else:
            x, y = self.x, self.y
        return x + at[0], y + at[1]

    def draw_branches(self, at=(0, 0), ax=None, symbol_nomenclature=None,
                      label=True, visited=None, **kwargs):
        '''
        Draw the edges linking parent nodes to child nodes. Also draws the parent outlink position
        and the child anomer symbol
        Parameters
        ----------
        at: :class:`tuple`
            The x, y coordinates at which to draw the glycan's reducing_end. Defaults to `(0, 0)`
        ax: :class:`matplotlib.axes.Axis`
            A matplotlib axis on which to draw. Useful for adding glycans to existing figures. If
            an axis is not provided, a new figure will be created and used.
        symbol_nomenclature: |str|
            A string mapping to the symbol nomenclature to use. Defaults to |None|. When `symbol_nomenclature`
            is |None|, `snfg` will be used.
        visited: |set|
            A |set| instance containing the node ids already visited while drawing. If the current
            node was already visited, its id will already be present. Used to avoid infinite cycles.
        '''
        if visited is None:
            visited = set()
        x, y = self.coords(at)
        if self.id in visited:
            return x, y
        visited.add(self.id)
        edge_weight = kwargs.get("edge_weight", 0.5)
        for child in self:
            cx, cy = child.draw_branches(at=at, ax=ax,
                                         symbol_nomenclature=symbol_nomenclature,
                                         label=label, visited=visited, **kwargs)
            patch = symbol_nomenclature.line_to(
                ax, x, y, cx, cy,
                color=kwargs.get('link_color', symbol_nomenclature.default_line_color),
                zorder=1, lw=edge_weight)
            self.data['lines'][self.id, child.id] = [patch]
            self.lines.append(patch)
        return x, y

    def draw_nodes(self, at=(0, 0), ax=None, symbol_nomenclature=None, label=True, visited=None, **kwargs):
        '''
        Draw the symbol representing the individual monosaccharide and its substituents.

        Parameters
        ----------
        at: :class:`tuple`
            The x, y coordinates at which to draw the glycan's reducing_end. Defaults to `(0, 0)`
        ax: :class:`matplotlib.axes.Axis`
            A matplotlib axis on which to draw. Useful for adding glycans to existing figures. If
            an axis is not provided, a new figure will be created and used.
        symbol_nomenclature: |str|
            A string mapping to the symbol nomenclature to use. Defaults to |None|. When `symbol_nomenclature`
            is |None|, `cfg` will be used.
        visited: |set|
            A |set| instance containing the node ids already visited while drawing. If the current
            node was already visited, its id will already be present. Used to avoid infinite cycles.
        '''
        if visited is None:
            visited = set()
        if self.id in visited:
            return
        visited.add(self.id)
        x, y = self.coords(at)
        patch_node = symbol_nomenclature.draw(self.tree, x, y, ax, tree_node=self,
                                              **kwargs)
        self.patch_node = patch_node
        residue_elements, substituent_elements, saccharide_label = patch_node
        self.data["patches"][self.tree.id] = [residue_elements]
        self.substituents_text = self.data["text"]["substituents"][self.id] = [substituent_elements]
        self.saccharide_label = self.data['text']['saccharide_label'][self.id] = [saccharide_label]
        self.data["position"][self.tree.id] = x, y

        patch_node.set_gid(self.uuid, self.id)

        for child in self:
            child.draw_nodes(at=at, ax=ax, symbol_nomenclature=symbol_nomenclature,
                             visited=visited, **kwargs)

    def _draw_linkage_label(self, child, at=(0, 0), ax=None, symbol_nomenclature=None, **kwargs):
        sx, sy = self.coords(at)
        cx, cy = child.coords(at)

        fontsize = kwargs.get("fontsize", 6)
        style = kwargs.get('label_style', 'split')

        position_x = (sx * 0.65 + cx * 0.35)
        if abs(sx - cx) < 1e-2:
            position_y = (sy + cy) / 2
        else:
            position_y = interpolate(sx, sy, cx, cy, position_x)

        position_num = -1
        for pos, link in self.tree.links.items():
            if link.is_child(child.tree):
                position_num = pos
                break

        anomer_x = (sx * 0.35 + cx * 0.65)
        if abs(sx - cx) < 1e-2:
            anomer_y = (sy + cy) / 2
        else:
            anomer_y = interpolate(sx, sy, cx, cy, anomer_x)

        if abs(anomer_x - position_x) < 1e-2 and abs(anomer_y - position_y) < 1e-2:
            anomer_x -= 0.07
            position_x += 0.05
            anomer_y -= 0.03
            position_y -= 0.03

        if style == 'split':
            position_text = symbol_nomenclature.draw_text(
                ax=ax, x=position_x, y=position_y + 0.03,
                text=str(position_num) if position_num != -1 else "?",
                fontsize=fontsize, ha='center', va='bottom')
            anomer_text = symbol_nomenclature.draw_text(
                ax, anomer_x, anomer_y + 0.03,
                r'${}$'.format(anomer_symbol_map.get(
                    child.tree.anomer, child.tree.anomer.name)),
                fontsize=fontsize, ha='center', va='bottom')
            self.data['text'][self.id, child.id]['linkage'] = [position_text, anomer_text]
        elif style == 'centered':
            symbol_nomenclature.draw_text(
                ax=ax, x=(position_x + anomer_x) / 2.0, y=(position_y + anomer_y) / 2.0,
                text=(str(position_num) if position_num != -1 else "?") + r'${}$'.format(anomer_symbol_map.get(
                    child.tree.anomer, child.tree.anomer.name)),
                fontsize=fontsize, ha='center', va='bottom')
        else:
            raise ValueError(
                "Unrecognized label_style position style %r" % (style, ))

    def draw_linkage_annotations(self, at=(0, 0), ax=None, symbol_nomenclature=None, child=None, **kwargs):
        '''
        Draw the parent outlink position and the child anomer symbol

        Parameters
        ----------
        at: :class:`tuple`
            The x, y coordinates at which to draw the glycan's reducing_end. Defaults to `(0, 0)`
        ax: :class:`matplotlib.axes.Axis`
            A matplotlib axis on which to draw. Useful for adding glycans to existing figures. If
            an axis is not provided, a new figure will be created and used.
        symbol_nomenclature: |str|
            A string mapping to the symbol nomenclature to use. Defaults to |None|. When `symbol_nomenclature`
            is |None|, `cfg` will be used.
        child: :class:`DrawTreeNode`
            The child node to draw relative to.
        '''
        if child is None:
            for child in self:
                self.draw_linkage_annotations(
                    at=at, ax=ax, symbol_nomenclature=symbol_nomenclature,
                    child=child, **kwargs)
                child.draw_linkage_annotations(
                    at=at, ax=ax, symbol_nomenclature=symbol_nomenclature, **kwargs)
            return
        self._draw_linkage_label(child, at, ax, symbol_nomenclature, **kwargs)

    # Tree Layout Helpers
    def left(self):
        return self.thread or len(self.children) and self.children[0]

    def right(self):
        return self.thread or len(self.children) and self.children[-1]

    def lbrother(self):
        n = None
        if self.parent:
            for node in self.parent.children:
                if node == self:
                    return n
                else:
                    n = node
        return n

    @property
    def lmost_sibling(self):
        if not self._lmost_sibling and self.parent and self != self.parent.children[0]:
            self._lmost_sibling = self.parent.children[0]
        return self._lmost_sibling

    def __str__(self):  # pragma: no cover
        return "%s: id=%s  x=%s mod=%s" % (self.tree, self.id, self.x, self.mod)

    def __repr__(self):  # pragma: no cover
        return self.__str__()

    def extrema(self, at=(0, 0), xmin=INF, xmax=-INF, ymin=INF, ymax=-INF, visited=None):
        '''
        Finds the most extreme points describing the area this node and its children occupy.

        Parameters
        ----------
        at: tuple of int
            The position to orient relative to in the xy-plane. A tuple of integers corresponding
            to the center x and y respectively.
        xmin: float
        xmax: float
        ymin: float
        ymax: float
            Used internally to pass along current extreme values recursively to child nodes.

        Returns
        -------
        xmin: float
        xmax: float
        ymin: float
        ymax: float

        '''
        if visited is None:
            visited = set()
        if self.id in visited:
            return xmin, xmax, ymin, ymax
        visited.add(self.id)
        x, y = self.coords(at)
        if x < xmin:
            xmin = x
        if x > xmax:
            xmax = x
        if y < ymin:
            ymin = y
        if y > ymax:
            ymax = y
        for child in sorted(self.children, key=lambda x: len(x.children)):
            n_xmin, n_xmax, n_ymin, n_ymax = child.extrema(at, xmin, xmax, ymin, ymax, visited=visited)
            xmin = min(xmin, n_xmin)
            ymin = min(ymin, n_ymin)
            xmax = max(xmax, n_xmax)
            ymax = max(ymax, n_ymax)
        return xmin, xmax, ymin, ymax

    def get(self, node_id):
        return get(self, node_id)

    def get_link_pair(self, link_id):
        return get_link_pair(self, link_id)

    def annotate_fragment(self, fragment, ax=None, color='red', label=True, **kwargs):
        if ax is None:
            ax = self.axes
        return BondCleavageArtist(self, fragment, ax, color=color, label=label, **kwargs)

    def set_transform(self, transform):
        """Apply the passed Affine2D to each graphical unit. The transform is additive
        with any previous affine transformation.

        where `M` is the value of `transform.get_matrix()`

        Parameters
        ----------
        transform : matplotlib.transforms.Affine2D
        """
        current_transform = self.transform
        if current_transform is not None:
            transform = transform + current_transform
        self.data["transform"] = transform

        # NOTE: Merge the aggregated transform of the graph
        # structure with the transformation induced by the Axes object
        # which draws each component. If this is skipped, all of the
        # geometries are wrong and the drawing will fail (or not even appear)
        transform = transform + self.axes.transData
        for i, patch_set in self.data['patches'].items():
            for entity in patch_set:
                for p in entity:
                    if isinstance(p, tuple):
                        [_.set_transform(transform) for _ in p]
                    else:
                        p.set_transform(transform)
        for i, patch_set in self.data['lines'].items():
            for p in patch_set:
                if isinstance(p, tuple):
                    [_.set_transform(transform) for _ in p]
                else:
                    p.set_transform(transform)
        for i, groups in self.data['text'].items():
            text = flatten(groups.values())
            [t.set_transform(transform) for t in text]

    def get_transform(self):
        if "transform" in self.data:
            trans = self.data["transform"]
            return trans
        else:
            return None

    transform = property(get_transform, set_transform)

    def _compute_bounding_box(self, pad_factor=0.2):
        xmin, xmax, ymin, ymax = self.extrema()
        dx = (max(xmin, xmax) - min(xmin, xmax)) * pad_factor
        dy = (max(ymin, ymax) - min(ymin, ymax)) * pad_factor
        return (xmin - dx, xmax + dx), (ymin - dy, ymax + dy)

    def update_text_position(self, degrees=0, preserve_orientation=True):
        '''
        This is a hack needed to orient text appropriately when the tree is rotated after labels
        are inscribed, as any new transform will override the correction.
        '''
        current_transform = self.transform
        if current_transform is None:
            current_transform = mtransforms.Affine2D()
        for i, groups in self.data['text'].items():
            for t in flatten(groups.values()):
                cent = (centroid(t.get_path().transformed(current_transform)))
                trans = current_transform.frozen().get_affine().translate(0, 0).rotate_deg(degrees=degrees)
                new_cent = centroid(t.get_path().transformed(trans))
                shift = cent - new_cent
                trans.translate(*shift)
                trans += self.axes.transData
                t.set_transform(trans)


class DrawTree(object):  # pragma: no cover
    def __init__(self, structure, figure=None, ax=None, layout=BalancedTreeLayout,
                 symbol_nomenclature=snfg_symbols.SNFGNomenclature(), **kwargs):
        self.structure = structure
        self.root = DrawTreeNode(root(structure))
        self.figure = figure
        self.ax = ax
        self.layout_scheme = layout
        self.symbol_nomenclature = symbol_nomenclature

    def layout(self, **kwargs):
        self.layout_scheme(self.root).layout()
        self.root.fix_special_cases()

    def draw(self, ax=None, center=True, **kwargs):
        if ax is not None:
            self.ax = ax
            self.figure = ax.get_figure()
        else:
            self.figure, self.ax = plt.subplots()

        at = (0, 0)

        self.root.draw(ax=self.ax, at=at, symbol_nomenclature=self.symbol_nomenclature)
        if center:
            xmin, xmax, ymin, ymax = self.root.extrema(at=at)
            ax = self.ax
            ax.set_xlim(sign(xmin) * (abs(xmin) + 2), (1 * abs(xmax) + 2))
            ax.set_ylim(sign(ymin) * (abs(ymin) + 2), (1 * abs(ymax) + 2))
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)

    def get(self, node_id):
        return self.root.get(node_id)

    def get_link_pair(self, link_id):
        return self.root.get_link_pair(link_id)

    @property
    def transform(self):
        return self.root.transform

    def run(self):
        self.layout()
        self.draw(ax=self.ax)
        self.figure.canvas.draw()

    def redraw(self, **kwargs):
        self.draw(**kwargs)
        self.figure.canvas.draw()


def plot(tree, at=(0, 0), ax=None, orientation='h', center=False, label=False,
         symbol_nomenclature='snfg', layout='balanced', layout_args=None, **kwargs):
    '''
    Draw the parent outlink position and the child anomer symbol

    Parameters
    ----------
    tree: Glycan or Monosaccharide
        The saccharide structure to draw.
    orientation: str
        A string corresponding to `h` or `horizontal` will draw the glycan horizontally
        from right to left. `v` or `vertical` will draw the glycan from bottom to top.
        Defaults to `h`
    at: tuple
        The x, y coordinates at which to draw the glycan's reducing_end. Defaults to `(0, 0)`
    ax: :class:`matplotlib.axes.Axis`
        A matplotlib axis on which to draw. Useful for adding glycans to existing figures. If
        an axis is not provided, a new figure will be created and used.
    center: bool
        Should the plot limits be centered around this glycan? Defaults to |False| but will be
        be set to |True| if `ax` is |None|
    label: bool
        Should the bond annotations for `tree` be drawn? Defaults to |False|
    symbol_nomenclature: str or :class:`~.SymbolicNomenclatureBase`
        Either the name of a symbol nomenclature or an instance of :class:`~.SymbolicNomenclatureBase`.
        Recognized names are "cfg", "snfg", and "iupac".
    layout: str or :class:`~.TreeLayoutBase`
        Either the name of a tree layout strategy or an instance of :class:`~.TreeLayoutBase`.
        Recognized names are "balanced" and "topological".
    layout_args: dict
        Extra parameters to pass to :class:`~.TreeLayoutBase`. It may take its own "transform" parameter
        passed here.
    scale: (float, float) or float
        Scale coefficients. Pass a pair to scale x and y dimensions respectively.
    symbol_scale: float, optional
        A scale coefficient for the size of the monosaccharide symbols, without changing the overall size
        of the layout.
    edge_weight: float, optional
        The line weight of the edges connecting monosaccharide symbols.
    label_style: str, optional
        The layout style used for drawing edge labels. May be one of "split" or "centered", where
        "split" will place the position close to the parent symbol and the anomer close to the
        child residue and "centered" places the symbols together at the centroid of the edge.
        Defaults to 'split'.
    '''

    scale = DEFAULT_SYMBOL_SCALE_FACTOR * kwargs.pop('symbol_scale', 1.0)
    transform_scale = kwargs.pop("scale", (1,))
    try:
        transform_scale = tuple(transform_scale)
    except Exception:
        transform_scale = (transform_scale, transform_scale)

    if isinstance(symbol_nomenclature, basestring):
        symbol_nomenclature = nomenclature_map.get(symbol_nomenclature)
    elif symbol_nomenclature is None:
        symbol_nomenclature = snfg_symbols.SNFGNomenclature()

    if isinstance(layout, basestring):
        layout = layout_map.get(layout)

    tree_root = root(tree)
    dtree = DrawTreeNode(tree_root)

    layout_algorithm = layout(dtree)
    layout_algorithm.layout(**(layout_args or {}))

    fig = None
    # Create a figure if no axes are provided
    if ax is None:
        fig, ax = plt.subplots()
        # at = (0, 0)
        center = True
        ax.axis('off')
    (layout_transform,
     substituent_transform,
     _) = symbol_nomenclature.get_layout_transform(orientation)
    layout_algorithm.transform(layout_transform)
    dtree.draw(at=at, ax=ax, scale=scale, label=False,
               symbol_nomenclature=symbol_nomenclature,
               annotation_transform=substituent_transform,
               **kwargs)
    dtree.axes = ax
    dtree.data['orientation'] = orientation
    dtree.set_transform(mtransforms.Affine2D())
    if label:
        dtree.draw_linkage_annotations(
            at=at, ax=ax, scale=scale,
            symbol_nomenclature=symbol_nomenclature,
            **kwargs)
    # If the figure is stand-alone, center it
    if fig is not None or center:
        xlim, ylim = _bounding_box_to_limits(dtree)
        ax.set_xlim(*xlim)
        ax.set_ylim(*ylim)
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
    return (dtree, ax)



'''The :mod:`glypy.io` module includes functions for reading and writing
:class:`~.Glycan` and :class:`~.Monosaccharide` objects in common text formats,
communicating with glycan structure databases, and interpreting glycan nomenclature.
'''


__all__ = [
    "glycoct", "glycoct_xml", "linear_code", "iupac",
    "glyspace", "wurcs", "monosaccharidedb",
    "format_constants_map",
    "nomenclature"
]



# cython: boundscheck=False

# Credit to Pyteomics - http://pythonhosted.org/pyteomics - for majority of design
import re
from .mass_dict import nist_mass
from .base import ChemicalCompositionError, composition_factory

cimport cython
from cpython cimport PY_MAJOR_VERSION

from cpython.ref cimport PyObject
from cpython.dict cimport (PyDict_GetItem, PyDict_SetItem, PyDict_Next,
                           PyDict_Keys, PyDict_Update, PyDict_DelItem, PyDict_Size)
from cpython.int cimport PyInt_AsLong, PyInt_Check, PyInt_FromLong


# if PY_MAJOR_VERSION < 3:
#     from cpython.string cimport PyString_Format

# cdef extern from *:
#     unicode PyUnicode_Format(object format, object args)

from glypy.composition.compat cimport PyStr_Format

from cpython.float cimport PyFloat_AsDouble
from cpython.tuple cimport PyTuple_GetItem
from cpython.list cimport PyList_GET_ITEM

# Forward Declaration
cdef: 
    str _atom = r'([A-Z][a-z+]*)(?:\[(\d+)\])?([+-]?\d+)?'
    str _formula = r'^({})*$'.format(_atom)
    str _isotope_string = r'^([A-Z][a-z+]*)(?:\[(\d+)\])?$'

    object isotope_pattern = re.compile(_isotope_string)
    object formula_pattern = re.compile(_formula)


@cython.boundscheck(False)
cdef str _parse_isotope_string(str label, int* isotope_num):
    cdef:
        # int isotope_num = 0
        int i = 0
        int in_bracket = False
        # str element_name
        str current
        list name_parts = []
        list num_parts = []
        #Isotope result
    for i in range(len(label)):
        current = label[i]
        if in_bracket:
            if current == "]":
                break
            num_parts.append(current)
        elif current == "[":
            in_bracket = True
        else:
            name_parts.append(current)
    element_name = (''.join(name_parts))
    if len(num_parts) > 0:
        isotope_num[0] = (int(''.join(num_parts)))
    else:
        isotope_num[0] = 0
    return element_name

def parse_isotope_string(str label):
    cdef:
        int iso
        str elem
    elem = _parse_isotope_string(label, &iso)
    return (elem, iso)

cdef str _make_isotope_string(str element_name, int isotope_num):
    """Form a string label for an isotope."""
    cdef:
        tuple parts
    if isotope_num == 0:
        return element_name
    else:
        parts = (element_name, isotope_num)
        return <str>PyStr_Format('%s[%d]', parts)


cdef class CComposition(dict):
    """A Composition object stores a chemical composition of a
    substance. Basically it is a dict object, in which keys are the names
    of chemical elements and values contain integer numbers of
    corresponding atoms in a substance.

    The main improvement over dict is that Composition objects allow
    addition and subtraction.

    If ``formula`` is not specified, the constructor will look at the first
    positional argument and try to build the object from it. Without
    positional arguments, a Composition will be constructed directly from
    keyword arguments.

    Parameters
    ----------
    formula : str, optional
        A string with a chemical formula. All elements must be present in
        `mass_data`.
    mass_data : dict, optional
        A dict with the masses of chemical elements (the default
        value is :py:data:`nist_mass`). It is used for formulae parsing only.
        """
    def __str__(self):   # pragma: no cover
        return 'Composition({})'.format(dict.__repr__(self))

    def __repr__(self):  # pragma: no cover
        return str(self)

    cdef CComposition add_from(self, other):
        cdef:
            str elem
            long cnt
            PyObject *pkey
            PyObject *pvalue
            Py_ssize_t ppos = 0

        while(PyDict_Next(other, &ppos, &pkey, &pvalue)):
            elem = <str>pkey
            cnt = self.getitem(elem)
            self.setitem(elem, cnt + PyInt_AsLong(<object>pvalue))

        self._mass_args = None
        return self

    def __iadd__(CComposition self, other):
        self.add_from(other)
        return self


    def __add__(self, other):
        cdef:
            CComposition result
        if not isinstance(self, CComposition):
            other, self = self, other
        result = CComposition._create(self)
        result.add_from(other)
        return result

    cdef CComposition subtract_from(self, other):
        cdef:
            str elem
            long cnt
            PyObject *pkey
            PyObject *pvalue
            Py_ssize_t ppos = 0

        while(PyDict_Next(other, &ppos, &pkey, &pvalue)):
            elem = <str>pkey
            cnt = self.getitem(elem)
            self.setitem(elem, cnt - PyInt_AsLong(<object>pvalue))

        self._mass_args = None
        return self

    def __isub__(self, other):
        return self.subtract_from(other)

    def __sub__(self, other):
        cdef:
            CComposition result
        if not isinstance(self, CComposition):
            self = CComposition(self)
        result = CComposition._create(self)
        result.subtract_from(other)
        return result

    def __reduce__(self):
        return composition_factory, (list(self),), self.__getstate__()

    def __getstate__(self):
        return dict(self)

    def __setstate__(self, d):
        self._from_dict(d)
        self._mass = None
        self._mass_args = None

    def __mul__(self, other):
        cdef:
            CComposition prod
            int rep, v
            str k
            PyObject *pkey
            PyObject *pvalue
            Py_ssize_t ppos = 0

        if isinstance(other, CComposition):
            self, other = other, self
        
        # prod = self.clone()
        # prod = CComposition()
        prod = CComposition._create(None)

        if not isinstance(other, int):
            raise ChemicalCompositionError(
                'Cannot multiply Composition by non-integer', other)

        rep = PyInt_AsLong(other)
        while(PyDict_Next(self, &ppos, &pkey, &pvalue)):
            k = <str>pkey
            v = PyInt_AsLong(<object>pvalue)
            prod.setitem(k, v * rep)

        return prod

    def __imul__(self, other):
        cdef:
            str k
            int v, rep
        if not isinstance(other, int):
            raise ChemicalCompositionError(
                'Cannot multiply Composition by non-integer',
                other)
        rep = other
        for k, v in list(self.items()):
            self.setitem(k, v * rep)
        return self

    def __eq__(self, other):
        cdef:
            Py_ssize_t self_size, other_size
            PyObject *pkey
            PyObject *pvalue
            PyObject *pinterm
            long self_value, other_value
        if not isinstance(other, dict):
            return False
        self_size = PyDict_Size(self)
        other_size = PyDict_Size(other)

        # if one dict is larger, it may contain more keys which may be zero,
        # which will be ignored.
        if other_size > self_size:
            other_size = 0
            while(PyDict_Next(other, &other_size, &pkey, &pvalue)):
                other_value = PyInt_AsLong(<object>pvalue)
                pinterm = PyDict_GetItem(self, <object>pkey)
                if pinterm == NULL:
                    self_value = 0
                else:
                    self_value = PyInt_AsLong(<object>pinterm)
                if self_value != other_value:
                    return False
        else:
            self_size = 0
            while(PyDict_Next(self, &self_size, &pkey, &pvalue)):
                self_value = PyInt_AsLong(<object>pvalue)
                pinterm = PyDict_GetItem(other, <object>pkey)
                if pinterm == NULL:
                    other_value = 0
                else:
                    other_value = PyInt_AsLong(<object>pinterm)
                if self_value != other_value:
                    return False
        return True

    def __ne__(self, other):
        cdef:
            Py_ssize_t self_size, other_size
            PyObject *pkey
            PyObject *pvalue
            PyObject *pinterm
            long self_value, other_value
        if not isinstance(other, dict):
            return True
        self_size = PyDict_Size(self)
        other_size = PyDict_Size(other)
        if other_size > self_size:
            other_size = 0
            while(PyDict_Next(other, &other_size, &pkey, &pvalue)):
                other_value = PyInt_AsLong(<object>pvalue)
                pinterm = PyDict_GetItem(self, <object>pkey)
                if pinterm == NULL:
                    self_value = 0
                else:
                    self_value = PyInt_AsLong(<object>pinterm)
                if self_value != other_value:
                    return True
        else:
            self_size = 0
            while(PyDict_Next(self, &self_size, &pkey, &pvalue)):
                self_value = PyInt_AsLong(<object>pvalue)
                pinterm = PyDict_GetItem(other, <object>pkey)
                if pinterm == NULL:
                    other_value = 0
                else:
                    other_value = PyInt_AsLong(<object>pinterm)
                if self_value != other_value:
                    return True
        return False

    def __neg__(self):
        return self * -1

    @staticmethod
    cdef CComposition _create(CComposition inst):
        cdef:
            CComposition result

        result = CComposition.__new__(CComposition)
        result._mass = None
        result._mass_args = None
        if inst is None:
            return result
        PyDict_Update(result, inst)
        return result


    # Override the default behavior, if a key is not present
    # do not initialize it to 0.
    def __missing__(self, str key):
        return 0

    def __setitem__(self, str key, object value):
        cdef long int_value = PyInt_AsLong(round(value))
        if int_value:  # Will not occur on 0 as 0 is falsey AND an integer
            self.setitem(key, int_value)
        elif key in self:
            del self[key]
        self._mass_args = None

    def copy(self):
        cdef CComposition dup
        dup = CComposition._create(None)
        PyDict_Update(dup, self)
        return dup

    cdef long getitem(self, str elem):
        cdef:
            PyObject* resobj
            long count
        resobj = PyDict_GetItem(self, elem)
        if (resobj == NULL):
            return 0
        count = PyInt_AsLong(<object>resobj)
        return count

    cdef void setitem(self, str elem, long val):
        PyDict_SetItem(self, elem, val)
        self._mass_args = None

    cpdef CComposition clone(self):
        return CComposition._create(self)

    def update(self, *args, **kwargs):
        dict.update(self, *args, **kwargs)
        self._mass_args = None

    @cython.boundscheck(False)
    cpdef _from_formula(self, str formula, dict mass_data):
        cdef:
            str elem, isotope, number
        if '(' in formula:
            self._from_formula_parens(formula, mass_data)
        elif not formula_pattern.match(formula):
            raise ChemicalCompositionError('Invalid formula: ' + formula)
        else:
            for elem, isotope, number in re.findall(_atom, formula):
                if not elem in mass_data:
                    raise ChemicalCompositionError('Unknown chemical element: ' + elem)
                self[_make_isotope_string(elem, int(isotope) if isotope else 0)
                        ] += int(number) if number else 1

    @cython.boundscheck(True)
    def _from_formula_parens(self, formula, mass_data):
        # Parsing a formula backwards.
        cdef:
            Py_ssize_t prev_chem_symbol_start, i
            int seek_mode, group_coef
            str parse_stack
            list resolve_stack


        prev_chem_symbol_start = len(formula)
        i = len(formula) - 1

        seek_mode = 0
        parse_stack = ""
        resolve_stack = []
        group_coef = 1

        while i >= 0:
            if seek_mode < 1:
                if (formula[i] == ")"):
                    seek_mode += 1
                    if i + 1 == prev_chem_symbol_start:
                        group_coef = 1
                    elif formula[i + 1].isdigit():
                        group_coef = int(formula[i + 1:prev_chem_symbol_start])
                    i -= 1
                    continue
                # Read backwards until a non-number character is met.
                if (formula[i].isdigit() or formula[i] == '-'):
                    i -= 1
                    continue

                else:
                    # If the number of atoms is omitted then it is 1.
                    if i + 1 == prev_chem_symbol_start:
                        num_atoms = 1
                    else:
                        try:
                            num_atoms = int(formula[i + 1:prev_chem_symbol_start])
                        except ValueError:
                            raise ChemicalCompositionError(
                                'Badly-formed number of atoms: %s' % formula)

                    # Read isotope number if specified, else it is undefined (=0).
                    if formula[i] == ']':
                        brace_pos = formula.rfind('[', 0, i)
                        if brace_pos == -1:
                            raise ChemicalCompositionError(
                                'Badly-formed isotope number: %s' % formula)
                        try:
                            isotope_num = int(formula[brace_pos + 1:i])
                        except ValueError:
                            raise ChemicalCompositionError(
                                'Badly-formed isotope number: %s' % formula)
                        i = brace_pos - 1
                    else:
                        isotope_num = 0

                    # Match the element name to the mass_data.
                    element_found = False
                    # Sort the keys from longest to shortest to workaround
                    # the overlapping keys issue
                    for element_name in sorted(mass_data, key=len, reverse=True):
                        if formula.endswith(element_name, 0, i + 1):
                            isotope_string = _make_isotope_string(
                                element_name, isotope_num)
                            self[isotope_string] += num_atoms
                            i -= len(element_name)
                            prev_chem_symbol_start = i + 1
                            element_found = True
                            break

                    if not element_found:
                        raise ChemicalCompositionError(
                            'Unknown chemical element in the formula: %s' % formula)
            else:
                ch = formula[i]
                parse_stack += ch
                i -= 1
                if(ch == "("):
                    seek_mode -= 1
                    if seek_mode == 0:

                        resolve_stack.append(Composition(
                                             # Omit the last character, then reverse the parse
                                             # stack string.
                                             formula=parse_stack[:-1][::-1],
                                             mass_data=mass_data)
                                             * group_coef)
                        prev_chem_symbol_start = i + 1
                        seek_mode = False
                        parse_stack = ""
                elif(formula[i] == ")"):
                    seek_mode += 1
                else:
                    # continue to accumulate tokens
                    pass

        # Unspool the resolve stack, adding together the chunks
        # at this level. __add__ operates immutably, so must manually
        # loop through each chunk.
        for chunk in resolve_stack:
            for elem, cnt in chunk.items():
                self[elem] += cnt

    cpdef _from_dict(self, comp):
        '''
        Directly overwrite this object's keys with the values in
        `comp` without checking their type.
        '''
        PyDict_Update(self, comp)

    cpdef double calc_mass(self, int average=False, charge=None, dict mass_data=nist_mass) except -1:
        cdef object mdid
        mdid = id(mass_data)
        if self._mass_args is not None and average is self._mass_args[0]\
                and charge == self._mass_args[1] and mdid == self._mass_args[2]:
            return self._mass
        else:
            self._mass_args = (average, charge, mdid)
            self._mass = calculate_mass(composition=self, average=average, charge=charge, mass_data=mass_data)
            return self._mass

    property mass:
        def __get__(self):
            return self.calc_mass()

    def __init__(self, args=None, formula=None, dict mass_data=None, **kwargs):
        #dict.__init__(self)
        if mass_data is None:
            mass_data = nist_mass

        if formula is not None:
            self._from_formula(formula, mass_data)

        elif args:
            if isinstance(args, dict):
                self._from_dict(args)
            elif isinstance(args, str):
                try:
                    self._from_formula(args, mass_data)
                except ChemicalCompositionError:
                    raise ChemicalCompositionError(
                        'Could not create a Composition object from '
                        'string: "{}": not a '
                        'formula'.format(args))
        else:
            self._from_dict(kwargs)
        self._mass = None
        self._mass_args = None

Composition = CComposition


cpdef CComposition composition_sum(list compositions):
    cdef:
        size_t i
        CComposition accumulator, current
        str elem
        long cnt
        PyObject *pkey
        PyObject *pvalue
        Py_ssize_t ppos = 0

    accumulator = CComposition()
    for i in range(len(compositions)):
        current = compositions[i]
        ppos = 0

        while(PyDict_Next(current, &ppos, &pkey, &pvalue)):
            elem = <str>pkey
            cnt = accumulator.getitem(elem)
            accumulator.setitem(elem, cnt + PyInt_AsLong(<object>pvalue))
    return accumulator


@cython.wraparound(False)
@cython.boundscheck(False)
cpdef double calculate_mass(CComposition composition=None, str formula=None, int average=False, charge=None, mass_data=None) except -1:
    """Calculates the monoisotopic mass of a chemical formula or CComposition object.

    Parameters
    ----------
    composition : CComposition
        A Composition object with the elemental composition of a substance. Exclusive with `formula`
    formula: str
        A string describing a chemical composition. Exclusive with `composition`
    average : bool, optional
        If :py:const:`True` then the average mass is calculated. Note that mass
        is not averaged for elements with specified isotopes. Default is
        :py:const:`False`.
    charge : int, optional
        If not 0 then m/z is calculated: the mass is increased
        by the corresponding number of proton masses and divided
        by z.
    mass_data : dict, optional
        A dict with the masses of the chemical elements (the default
        value is :py:data:`nist_mass`).

    Returns
    -------
        mass : float
    """
    cdef:
        int old_charge, isotope_num, isotope, quantity
        double mass, isotope_mass, isotope_frequency
        long _charge
        str isotope_string, element_name
        dict mass_provider
        list key_list
        PyObject* interm
        Py_ssize_t iter_pos = 0

    if mass_data is None:
        mass_provider = nist_mass
    else:
        mass_provider = mass_data

    if composition is None:
        if formula is not None:
            composition = CComposition(formula)
        else:
            raise ChemicalCompositionError("Must provide a composition or formula argument")
    else:
        if formula is not None:
            raise ChemicalCompositionError("Must provide a composition or formula argument, but not both")

    # Get charge.
    if charge is None:
        charge = composition.getitem('H+')
    else:
        if charge != 0 and composition.getitem('H+') != 0:
            raise ChemicalCompositionError("Charge is specified both by the number of protons and parameters")
    _charge = PyInt_AsLong(charge)
    old_charge = composition.getitem('H+')
    composition.setitem('H+', charge)

    # Calculate mass.
    mass = 0.0
    key_list = PyDict_Keys(composition)
    for iter_pos in range(len(key_list)):
        isotope_string = <str>PyList_GET_ITEM(key_list, iter_pos)
        # element_name, isotope_num = _parse_isotope_string(isotope_string)
        element_name = _parse_isotope_string(isotope_string, &isotope_num)

        # Calculate average mass if required and the isotope number is
        # not specified.
        if (not isotope_num) and average:
            for isotope in mass_provider[element_name]:
                if isotope != 0:
                    quantity = <int>composition.getitem(element_name)
                    isotope_mass = <double>mass_provider[element_name][isotope][0]
                    isotope_frequency = <double>mass_provider[element_name][isotope][1]

                    mass += quantity * isotope_mass * isotope_frequency
        else:
            interim = PyDict_GetItem(mass_provider, element_name)
            interim = PyDict_GetItem(<dict>interim, isotope_num)
            isotope_mass = PyFloat_AsDouble(<object>PyTuple_GetItem(<tuple>interim, 0))

            mass += (composition.getitem(isotope_string) * isotope_mass)

    # Calculate m/z if required.
    if _charge != 0:
        mass /= abs(_charge)

    if old_charge != 0:
        composition.setitem('H+', old_charge)
    else:
        PyDict_DelItem(composition, "H+")
    return mass



import multiprocessing

from collections import defaultdict

from glypy.algorithms import DistinctGlycanSet

from .graph import _enzyme_graph_inner


class Glycome(object):

    def __init__(self, glycosylases, glycosyltransferases, seeds, track_generations=False,
                 limits=None):
        if limits is None:
            limits = []
        self.glycosylases = glycosylases
        self.glycosyltransferases = glycosyltransferases
        self.seeds = seeds

        self.seen = DistinctGlycanSet()
        self.track_generations = track_generations
        self.enzyme_graph = defaultdict(_enzyme_graph_inner)
        self.history = []
        self.current_generation = DistinctGlycanSet(seeds)
        self.limits = limits

    def save_generation(self, generation):
        if self.track_generations:
            self.history.append(generation)
        self.seen.update(generation)

    def run(self, n=50):
        for i in range(n):
            generation = self.step()
            if not generation:
                break
            yield generation

    def clean_next_generation(self, generation):
        return generation - self.seen

    def within_limits(self, structure):
        for limiter in self.limits:
            if not limiter(structure):
                return False
        return True

    def step(self):
        next_generation = DistinctGlycanSet()
        for species in self.current_generation:
            parentkey = None
            for enzkey, enz in self.glycosylases.items():
                products = [root for root, leaf in enz(
                    species, refund=True) if self.within_limits(root)]
                if products:
                    if parentkey is None:
                        parentkey = str(species)
                    for product in products:
                        childkey = str(product)
                        self.enzyme_graph[parentkey][childkey].add(enzkey)
                        next_generation.add(product)
            for enzkey, enz in self.glycosyltransferases.items():
                products = [root for root in enz(
                    species) if self.within_limits(root)]
                if products:
                    if parentkey is None:
                        parentkey = str(species)
                    for product in products:
                        childkey = str(product)
                        self.enzyme_graph[parentkey][childkey].add(enzkey)
                        next_generation.add(product)
        self.save_generation(self.current_generation)
        self.current_generation = self.clean_next_generation(next_generation)
        return next_generation


def _MultiprocessingGlycome_worker(seeds_params):
    seeds, params, seen = seeds_params
    import dill
    (glycosylases, glycosyltransferases, _,
     track_generations, limits) = dill.loads(params)
    # limits.append(lambda x: x not in seen)
    glycome = Glycome(glycosylases, glycosyltransferases, seeds,
                      track_generations, limits)
    glycome.step()
    return glycome.current_generation, glycome.enzyme_graph


class MultiprocessingGlycome(Glycome):

    def __init__(self, glycosylases, glycosyltransferases, seeds, track_generations=False,
                 limits=None, processes=None):
        if processes is None:
            processes = min(multiprocessing.cpu_count(), 4)
        super(MultiprocessingGlycome, self).__init__(
            glycosylases, glycosyltransferases, seeds,
            track_generations, limits)
        self.processes = processes
        self.pool = None
        self.seen = DistinctGlycanSet()
        self._worker_params = (
            self.glycosylases, self.glycosyltransferases, tuple(),
            self.track_generations, self.limits)

    def _create_pool(self):
        self.pool = multiprocessing.Pool(self.processes)

    def _log(self, message):
        print(message)

    def log_generation_chunk(self, i, chunks, current_generation):
        self._log(".... Task %d/%d finished (%d items generated)" % (
            i, len(chunks), len(current_generation)))

    def _generate_work_loads(self, next_generation, chunks):
        import dill
        encoded_params = dill.dumps(self._worker_params)

        for i, chunk in enumerate(chunks):
            yield (chunk, encoded_params, next_generation)

    def _partition_generation(self, generation, max_chunk_size=2e3):
        n = len(generation)
        n_chunks = int(n // max_chunk_size)
        if n_chunks < self.processes * 4:
            return generation.partition(self.processes * 4)
        else:
            return generation.partition(n_chunks)

    def step(self):
        next_generation = DistinctGlycanSet()
        self._log(".... Starting Step")
        chunks = self._partition_generation(self.current_generation)
        self._log(".... Produced %d chunks" % (len(chunks) + 1,))
        work_spec = self._generate_work_loads(next_generation, chunks)
        if self.pool is None:
            self._create_pool()
        pool = self.pool
        i = 0
        for work in pool.imap_unordered(_MultiprocessingGlycome_worker, work_spec):
            i += 1
            current_generation, enzyme_graph = work
            self.log_generation_chunk(i, chunks, current_generation)
            next_generation.update(current_generation)
            for parent, children in enzyme_graph.items():
                for child, enzymes in children.items():
                    self.enzyme_graph[parent][child].update(enzymes)

        self.save_generation(self.current_generation)
        self.current_generation = self.clean_next_generation(next_generation)
        return next_generation



import re
from glypy.utils import opener, Enum, root, StringIO
from glypy.structure.glycan import NamedGlycan
from six import add_metaclass


class ParserState(Enum):
    defline = 1
    sequence = 2


class ParserError(ValueError):
    pass


class FormatRegisteringMeta(type):
    def __new__(cls, name, parents, attrs):
        new_type = type.__new__(cls, name, parents, attrs)
        if not hasattr(cls, "registry"):
            cls.registry = {}

        if 'format_name' in attrs:
            cls.registry[attrs["format_name"]] = new_type
        return new_type


class TextFileParserBase(object):

    def __init__(self, *args, **kwargs):
        self._iter = None

    def parse(self):  # pragma: no cover
        raise NotImplementedError()

    def __iter__(self):
        self._iter = self.parse()
        return self._iter

    def __next__(self):
        if self._iter is None:
            iter(self)
        return next(self._iter)

    next = __next__

    @classmethod
    def loads(cls, text):
        return cls(StringIO(text))


@add_metaclass(FormatRegisteringMeta)
class FastaLikeFileParser(TextFileParserBase):
    format_name = 'fasta'

    def __init__(self, file_spec, processor):
        super(FastaLikeFileParser, self).__init__()
        self.state = ParserState.defline
        self.handle = opener(file_spec)
        self.defline = None
        self.sequence_chunks = []
        self.processor = processor

    def parse(self):
        for line in self.handle:
            line = line.lstrip()
            if self.state is ParserState.defline:
                if line[0] == ">":
                    self.defline = re.sub(r"[\n\r]", "", line[1:])
                    self.state = ParserState.sequence
                else:
                    continue
            else:
                if not re.match(r"^(\s+|>)", line):
                    self.sequence_chunks.append(re.sub(r"[\n\r]", "", line))
                else:
                    if self.defline is not None:
                        yield self.pack()

                    self.sequence_chunks = []
                    self.defline = None
                    self.state = ParserState.defline
                    if line[0] == '>':
                        self.defline = re.sub(r"[\n\r]", "", line[1:])
                        self.state = ParserState.sequence

        if len(self.sequence_chunks) > 0:
            yield self.pack()

    def pack(self):
        d = self.processor(''.join(self.sequence_chunks))
        name = self.defline
        root_node = root(d)
        return NamedGlycan(name=name, root=root_node, index_method='dfs')


@add_metaclass(FormatRegisteringMeta)
class StructurePerLineParser(TextFileParserBase):
    format_name = 'line'

    def __init__(self, file_spec, processor):
        super(StructurePerLineParser, self).__init__()
        self.handle = opener(file_spec)
        self.processor = processor

    def parse(self):
        for line in self.handle:
            data = self.processor(line)
            yield data


class ParserInterface(TextFileParserBase):

    def __init__(self, file_spec, file_type='line'):
        line_fetcher = FormatRegisteringMeta.registry[file_type]
        self.reader = line_fetcher(file_spec, processor=self.process_result)
        self.file_type = file_type

        super(ParserInterface, self).__init__()

    def parse(self):
        return self.reader.parse()

    @classmethod
    def loads(cls, text, file_type='line'):
        return cls(StringIO(text), file_type=file_type)

    @classmethod
    def load(cls, fd, file_type='line'):
        reader = cls(fd, file_type=file_type)
        data = list(reader)
        if len(data) == 1:
            return data[0]
        else:
            return data




class ChemicalCompositionError(Exception):
    pass


def composition_factory(*args, **kwargs):  # pragma: nocover
    '''
    Late-binding factory function. Used to reconstitute Composition objects
    that have been pickled, where the C-extension version may not be available
    on deserialization.
    '''
    try:
        from .composition import CComposition as Composition
    except ImportError:
        from .composition import PComposition as Composition
    return Composition()


def formula(composition):
    """Convert a :class:`~.Composition` into a :class:`str` form that can
    be parsed.

    Parameters
    ----------
    composition: :class:`~.Composition`
        A chemical composition

    Returns
    -------
    :class:`str`
    """
    return ''.join("%s%d" % (k, v) for k, v in sorted(composition.items()))



from functools import cmp_to_key
from collections import Counter, deque


def all_node_depth(node, visited=None):
    if visited is None:
        visited = set()
    if node.id in visited:  # pragma: no cover
        return 0
    visited.add(node.id)
    depth_count = 1
    children = list(node.children())
    try:
        children += list(node.substituents())
    except AttributeError:
        pass
    if children:
        depth_count += max(all_node_depth(ch, visited) for p, ch in children)
    return depth_count


class CanonicalizerBase(object):
    def __init__(self, structure, reverse=False):
        self.structure = structure
        self.reverse = reverse
        if not self.structure.has_index():
            self.structure.reindex()

    def sort_links(self, links, reverse=False):
        raise NotImplementedError()

    def resort(self):
        nodes = deque(self.structure.leaves())
        seen = set()
        while nodes:
            node = nodes.popleft()
            if node.id in seen:
                continue
            seen.add(node.id)
            link_map = node.links.__class__()
            for link in self.sort_links(node.links.values(), reverse=self.reverse):
                if link.is_parent(node):
                    link_map[link.parent_position] = link
                else:
                    link_map[link.child_position] = link
            node.links = link_map
            for pos, parent in node.parents():
                nodes.append(parent)
        for link in self.structure.link_index:
            link.label = None
        self.structure.label_branches()

    @classmethod
    def canonicalize(cls, structure, **kwargs):
        sorter = cls(structure, **kwargs)
        sorter.resort()
        return structure


class GlycoCTCanonicalizer(CanonicalizerBase):
    def __init__(self, structure, reverse=False):
        super(GlycoCTCanonicalizer, self).__init__(structure, reverse=reverse)
        self.branch_to_terminal_count = self.build_branch_to_terminal_count()

    def get_branch_from_link_label(self, link):
        return link.label[0]

    def build_branch_to_terminal_count(self):
        counter = Counter()
        try:
            for key in sorted(self.structure.branch_parent_map.keys(), reverse=True):
                parent = self.structure.branch_parent_map[key]
                counter[parent] += counter[key] + 1
        except AttributeError:
            pass
        return counter

    def _compare_residue_ordering(self, res_a, res_b):
        n_child_residues_a = all_node_depth(res_a)
        n_child_residues_b = all_node_depth(res_b)
        diff_child_res = n_child_residues_a - n_child_residues_b

        if diff_child_res != 0:
            if diff_child_res < 0:
                return -1
            else:
                return 1

        try:
            branch_length_a = max((all_node_depth(cr) for p, cr in res_a.children()))
        except ValueError:
            branch_length_a = 0
        try:
            branch_length_b = max((all_node_depth(cr) for p, cr in res_b.children()))
        except ValueError:
            branch_length_b = 0

        diff_longest_branch = branch_length_a - branch_length_b

        if diff_longest_branch != 0:
            if diff_longest_branch < 0:
                return -1
            else:
                return 1

        n_branches_from_a = 0
        n_branches_from_b = 0
        for link in res_a.links.values():
            if link.is_parent(res_a):
                branch_label = self.get_branch_from_link_label(link)
                n_branches_from_a = max(n_branches_from_a, self.branch_to_terminal_count[branch_label])

        for link in res_b.links.values():
            if link.is_parent(res_b):
                branch_label = self.get_branch_from_link_label(link)
                n_branches_from_b = max(n_branches_from_b, self.branch_to_terminal_count[branch_label])
        diff_n_branches_from = n_branches_from_a - n_branches_from_b

        if diff_n_branches_from != 0:
            if diff_n_branches_from < 0:
                return -1
            else:
                return 1

        if res_a == res_b:
            return 0

        subtree_a = str(self.structure.subtree_from(self.structure, res_a))
        subtree_b = str(self.structure.subtree_from(self.structure, res_b))
        return (subtree_b > subtree_a) - (subtree_b < subtree_a)

    def compare_residue_ordering(self, res_a, res_b):
        ordered = self._compare_residue_ordering(res_a, res_b)
        return ordered

    def _compare_link_ordering(self, link_a, link_b):
        # Ignoring # of links for now since it is difficult
        # to compute
        parent_pos_a = link_a.parent_position
        parent_pos_b = link_b.parent_position
        try:
            diff_parent = parent_pos_a - parent_pos_b
        except TypeError as e:
            print(parent_pos_a, parent_pos_b, link_a, link_b)
            raise e

        if diff_parent != 0:
            if diff_parent < 0:
                return -1
            else:
                return 1

        child_pos_a = link_a.child_position
        child_pos_b = link_b.child_position
        diff_child = child_pos_a - child_pos_b

        if diff_child != 0:
            if diff_child < 0:
                return -1
            else:
                return 1

        sigils_a = link_a._glycoct_sigils()
        sigils_b = link_b._glycoct_sigils()

        if sigils_a[0] != sigils_b[0]:
            diff_sig0 = ord(sigils_a[0]) - ord(sigils_b[0])
            if diff_sig0 < 0:
                return -1
            else:
                return 1

        if sigils_a[1] != sigils_b[1]:
            diff_sig1 = ord(sigils_a[1]) - ord(sigils_b[1])
            if diff_sig1 < 0:
                return -1
            else:
                return 1

        child_a = link_a.child
        child_b = link_b.child
        ordered = self.compare_residue_ordering(child_a, child_b)
        return ordered

    def compare_link_ordering(self, link_a, link_b):
        ordered = self._compare_link_ordering(link_a, link_b)
        return ordered

    def sort_links(self, links, reverse=False):
        return sorted(links, key=cmp_to_key(self.compare_link_ordering),
                      reverse=reverse)

    def sort_residues(self, residues, reverse=False):
        return sorted(residues, key=cmp_to_key(self.compare_residue_ordering),
                      reverse=reverse)


def canonicalize(structure, canonicalizer=GlycoCTCanonicalizer, **kwargs):
    if canonicalizer is None:
        canonicalizer = GlycoCTCanonicalizer
    return canonicalizer.canonicalize(structure, **kwargs)



from functools import total_ordering

from six import add_metaclass
try:
    intern
except NameError:
    from sys import intern # pylint: disable=no-name-in-module


class EnumValue(object):
    '''Represents a wrapper around an value with a name to identify it and
    more rich comparison logic. A value of an enumerated type'''

    __slots__ = ('group', 'name', 'value', 'names', '_hash')

    def __init__(self, group, name, value, other_names=None):
        self.name = intern(name)
        self.value = value
        self.names = {name} | (other_names or set())
        self.group = group
        self._hash = None

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self.name)
        return self._hash

    def __index__(self):  # pragma: no cover
        return self.value

    def __int__(self):
        return int(self.value)

    def __eq__(self, other):
        try:
            if self.group is not other.group:
                return False
            # if self is other:
            #     return True
            # return self.value == other.value or self.names == other.names
            return self is other
        except AttributeError:
            return self.value == other or other in self.names

    def __and__(self, other):
        return self.value & other

    def __or__(self, other):
        return self.value | other

    def __rand__(self, other):
        return self.value & other

    def __ror__(self, other):
        return self.value | other

    def __xor__(self, other):
        return self.value ^ other

    def __rxor__(self, other):
        return self.value ^ other

    def __ne__(self, other):
        return not self == other

    def __repr__(self):  # pragma: no cover
        return "<{group_name} {name}:{value}>".format(name=self.name,
                                                      group_name=self.group.__name__,
                                                      value=self.value)

    def __str__(self):
        return self.name

    def __reduce__(self):
        return self.group, (self.name,)

    def add_name(self, name, force=False):
        if name not in self.group or force:
            self.names.add(name)
            self.group[name] = self
        else:
            raise KeyError("{} already exists in {}".format(name, self.group))

    def resolve(self, mapping):
        for name in self.names:
            try:
                if name in mapping:
                    return mapping[name]
            except KeyError:  # pragma: no cover
                pass
        raise KeyError("Could not resolve {} against {}".format(self, mapping))

    def __lt__(self, other):
        return self.value < other.value

    def __gt__(self, other):
        return self.value > other.value

    def __le__(self, other):
        return self.value <= other.value

    def __ge__(self, other):
        return self.value >= other.value

    def int_value(self):
        return (int(self.value))

try:
    _EnumValue = EnumValue
    _has_c = True
    from glypy.utils.cenum import EnumValue
except ImportError:
    _has_c = False


debug = False
QMARK = "?"


class EnumMeta(type):
    '''
    A metaclass for types hosting enumerated members. Class attributes are
    automatically translated into ``EnumValue`` objects with the specified value
    and a name string matching the specified attribute name. Newly assigned attributes
    are wrapped dynamically.

    The class itself can be treated like a dictionary for looking up these attributes
    by name or value, and these values can be iterated over.

    ..note::
        Why is this necessary? It's probably not. I chose to write it initially for
        compatibility with a previous iteration of the library, but later decided it
        was worth keeping for two reasons:

        1. Avoids ``stringly-typing`` the library. Comparison of strings is a slippery slope
           to raw magic strings littering the codebase.
        2. Richer comparison behavior allows these same names to be used in different modes with the
           same symbol.
        3. Namespacing of EnumValue objects makes it easier to avoid accidental name collisions
           when comparing EnumValues instead of magic strings.

    '''

    def __new__(cls, name, parents, attrs):
        if attrs.get('__doc__') is None:
            attrs['__doc__'] = "EnumType"
        enum_type = type.__new__(cls, name, parents, attrs)
        mapped = {}
        attr_pairs = list(attrs.items())
        EnumType = attrs.get("__enum_type__", EnumValue)
        for label, value in attr_pairs:
            if not label.startswith("__") or label == "mro":
                attrs.pop(label)
                delattr(enum_type, label)
                enum_value = EnumType(enum_type, label, value)
                if value in mapped:
                    try:
                        mapped[value].add_name(label)
                        setattr(enum_type, label, mapped[value])
                    except KeyError as e:
                        print(e)
                else:
                    mapped[value] = enum_value
                    setattr(enum_type, label, enum_value)
        enum_type[QMARK] = None
        for tp in parents:
            if isinstance(tp, EnumMeta):
                for label, value in tp:
                    if label == QMARK:
                        continue
                    super(EnumMeta, enum_type).__setattr__(label, value)
        return enum_type

    def __iter__(self):
        for attr, val in self.__dict__.items():
            if not attr.startswith("__") or attr == "mro":
                yield (attr, val)

    def __contains__(self, k):
        val = (k in self.__dict__) or (k in self.__dict__.values())
        return val

    def __getitem__(self, k):
        return self.translate(k)  # pylint: disable=no-value-for-parameter

    def __setattr__(self, k, v):
        """Intercept attribute assignment, wrapping values in
        :class:`EnumValue`

        Parameters
        ----------
        k : str
            Name to be set
        v : object
            The value to assign. If it is not of type :class:`EnumValue` it
            will be wrapped like `EnumValue(self, name=k, value=v)`
        """
        if isinstance(v, EnumValue):
            v.names.add(k)
            super(EnumMeta, self).__setattr__(k, v)
        else:
            name = self.name(v)  # pylint: disable=no-value-for-parameter
            if name is not None:
                self[name].add_name(k)  # pylint: disable=unsubscriptable-object
            else:
                super(EnumMeta, self).__setattr__(k, EnumValue(self, k, v))

    def __setitem__(self, k, v):
        setattr(self, k, v)

    def translate(self, k):
        '''
        Attempt to translate the input object ``k`` into a data member of the Enum.

        First try to find an element of ``self`` by hashing it against member names.

        Then try to find an element of ``self`` by searching for a member in self that
        is value-equal to ``k``

        Otherwise throw a :exc:`KeyError`

        Parameters
        ----------
        k: object
            The value to be translated.

        Returns
        -------
        :class:`EnumValue`

        '''
        # global debug
        # if debug:
        #     print "Translating", k

        if k in self.__dict__:
            return self.__dict__[k]
        elif k in self.__dict__.values():
            return self[self.name(k)]  # pylint: disable=unsubscriptable-object,no-value-for-parameter
        else:
            raise KeyError("Could not translate {0} through {1}".format(k, self))

    def name(self, v):
        for k, val in reversed(list(self)):
            if v == val:
                return k

    def __repr__(self):
        return "<Enum {0}>".format(self.__name__)

    __call__ = translate


try:
    _EnumMeta = EnumMeta
    _has_c = True
    from glypy.utils.cenum import EnumMeta
except ImportError:
    _has_c = False


@add_metaclass(EnumMeta)
class Enum(object):
    '''
    A simple class implementing :class:`EnumMeta`. Useful base type for other
    enumerated types.
    '''
    # __metaclass__ = EnumMeta

    def __init__(self):
        raise Exception("This class is not meant to be instantiated. Reference its attribute members directly")



from glypy.utils.enum import Enum


class SuperClass(Enum):
    '''Corresponds to the number of carbon atoms in the carbohydrate backbone
    of the |Monosaccharide|

    Is an |Enum|
    '''
    #: 2 carbons
    sug = 2
    #: Triose - 3 carbons
    tri = 3
    #: Tetrose - 4 carbons
    tet = 4
    #: Pentose - 5 carbons
    pen = 5
    #: Hexose - 6 carbons
    hex = 6
    #: Heptose - 7 carbons
    hep = 7
    #: Octose - 8 carbons
    oct = 8
    #: Nonose - 9 carbons
    non = 9
    #: Decose - 10 carbons
    dec = 10
    #: 11 carbons
    s11 = 11
    #: 12 carbons
    s12 = 12
    #: 13 carbons
    s13 = 13
    #: 14 carbons
    s14 = 14
    #: 15 carbons
    s15 = 15
    #: 16 carbons
    s16 = 16
    #: 17 carbons
    s17 = 17
    #: 18 carbons
    s18 = 18
    #: 19 carbons
    s19 = 19
    #: 20 carbons
    s20 = 20
    #: Unknown
    x = None


#: Sugose alias of `sug`
SuperClass.sug.add_name("Sugose")
#: Triose alias of `tri`
SuperClass.tri.add_name("Triose")
#: Tetrose alias of `tet`
SuperClass.tet.add_name("Tetrose")
#: Pentose alias of `pen`
SuperClass.pen.add_name("Pentose")
#: Hexose alias of `hex`
SuperClass.hex.add_name("Hexose")
#: Heptose alias of `hep`
SuperClass.hep.add_name("Heptose")
#: Octose alias of `oct`
SuperClass.oct.add_name("Octose")
#: Nonose alias of `non`
SuperClass.non.add_name("Nonose")
#: Decose alias of `dec`
SuperClass.dec.add_name("Decose")


class Stem(Enum):
    '''Corresponds to the bond formation pattern between the carbon atoms in the
    carbohydrate backbone of the |Monosaccharide|

    Is an |Enum|
    '''
    #: Glyceraldehyde
    gro = 1
    #: Erythrose
    ery = 2
    #: Ribose
    rib = 3
    #: Arabinose
    ara = 4
    #: Allose
    all = 5
    #: Altrose
    alt = 6
    #: Glucose
    glc = 7
    #: Mannose
    man = 8
    #: Threose
    tre = 9
    #: Xylose
    xyl = 10
    #: Lyxose
    lyx = 11
    #: Gulose
    gul = 12
    #: Idose
    ido = 13
    #: Galactose
    gal = 14
    #: Talose
    tal = 15
    #: Unknown
    x = None

    thr = 16


Stem.gro.add_name("Glyceraldehyde")
Stem.ery.add_name("Erythrose")
Stem.rib.add_name("Ribose")
Stem.ara.add_name("Arabinose")
Stem.all.add_name("Allose")
Stem.alt.add_name("Altose")
Stem.glc.add_name("Glucose")
Stem.man.add_name("Mannose")
Stem.tre.add_name("Threose")
Stem.xyl.add_name("Xylose")
Stem.lyx.add_name("Lyxose")
Stem.gul.add_name("Gulose")
Stem.ido.add_name("Idose")
Stem.gal.add_name("Galactose")
Stem.tal.add_name("Talose")
Stem.x.add_name("Unknown")


class Configuration(Enum):
    '''
    Corresponds to the optical stereomeric state of the |Monosaccharide|

    Is an |Enum|
    '''
    #: D Configuration
    d = 1
    #: L Configuration
    l = 2
    x = None


Configuration.d.add_name("Dextro")
Configuration.l.add_name("Levo")
Configuration.x.add_name("Unknown")


class Modification(Enum):
    '''
    Corresponds to discrete composition shifts of the |Monosaccharide| which
    are simple enough to not constitute a distinct object to represent like |Substituent|.

    Is an |Enum|
    '''
    #: Deoxygenated
    d = 1
    #: Ketone
    keto = 2
    #: DoubleBond
    en = 3
    #: Acidic
    a = 4
    #: Alditol
    aldi = 5
    #: SP2
    sp2 = 6
    #: SP
    sp = 7
    #: Geminal
    geminal = 8


#: alias of `aldi`
Modification.aldi.add_name("Alditol")
#: alias of `a`
Modification.a.add_name("Acidic")
#: alias of `d`
Modification.d.add_name("Deoxygenated")
#: alias of `keto`
Modification.keto.add_name("Ketone")
#: alias of `en`
Modification.en.add_name("DoubleBond")
#: alias of `geminal`
Modification.geminal.add_name("Geminal")
#: alias of `sp`
Modification.sp.add_name("SP")
#: alias of `sp2`
Modification.sp2.add_name("SP2")


class Anomer(Enum):
    '''
    Corresponds to the type of linkage found at the anomeric carbon of this |Monosaccharide|

    Is an |Enum|
    '''
    #: Alpha linkage
    alpha = 1
    #: Beta linkage
    beta = 2
    #: Uncyclized open chain
    uncyclized = 3
    #: Unknown
    x = None


Anomer.beta.add_name("b")
Anomer.alpha.add_name("a")
Anomer.uncyclized.add_name("o")
Anomer.uncyclized.add_name("open-chain")


class RingType(Enum):
    '''
    Corresponds to the type of ring structure of this |Monosaccharide|. Pyranose rings are
    five-member rings including one Oxygen and Furanose rings are four member rings including
    one Oxygen.

    Is an |Enum|
    '''
    #: Six member ring
    pyranose = 6
    #: Five member ring
    furanose = 5
    #: Open chain
    open = 0
    #: Unknown
    x = None


class Stereocoding(Enum):
    x = None
    h = 'h'
    L = '1'
    D = '2'
    LD = '3'
    DL = '4'
    d = 'd'
    m = 'm'
    a = 'a'
    o = 'o'
    k = 'k'
    e = 'e'
    n = 'n'
    E = 'E'
    y = 'y'
    s = 's'
    t = 't'


Stereocoding.h.add_name('0')


UnknownPosition = -1
NoPosition = None


class LinkageType(Enum):
    backbone_oxygen = 0
    backbone_hydrogen = 1
    other = 2
    unknown = None


LinkageType.unknown.add_name('x')



cimport cython


@cython.freelist(100000)
cdef class MoleculeBase(object):
    node_type = object()

    def __cinit__(self, *args, **kwargs):
        self._order = 0

    cpdef order(self, deep=False):
        '''
        Return the "graph theory" order of this molecule

        Returns
        -------
        int
        '''
        return self._order

    cpdef has_undefined_linkages(self):
        return True

    def copy(self, *args, **kwargs):
        return self.clone(*args, **kwargs)


cdef class SaccharideBase(MoleculeBase):
    node_type = object()


cdef class SaccharideCollection(SaccharideBase):
    cpdef _derivatized(self, substituent, id_base):
        pass

    cpdef _strip_derivatization(self):
        pass


cdef class SubstituentBase(MoleculeBase):
    node_type = object()


cdef class ModificationBase(MoleculeBase):
    node_type = object()


import pkg_resources
import logging
from rdflib import Graph, Namespace

logging.getLogger("rdflib.term").addHandler(logging.NullHandler())
logging.getLogger("rdflib.term").propagate = False


NSGlycan = ("http://purl.jp/bio/12/glyco/glycan#")
glycordf = Graph().parse(pkg_resources.resource_stream(__name__, "data/glycordf.ttl"), format='turtle')
_entities = set(glycordf.subjects()) | set(glycordf.objects())
_entities = [
    entity.replace(NSGlycan, "") for entity in _entities if (NSGlycan in entity)
]


class PopulatedNamespace(Namespace):
    def __new__(cls, value, members):
        inst = super(cls, cls).__new__(cls, value)
        inst._members = members
        return inst

    def __dir__(self):
        members = dir(super(PopulatedNamespace, self))
        members = sorted(set(list(members) + self._members))
        return members


NSGlycan = PopulatedNamespace(NSGlycan, _entities)


__all__ = [
    "NSGlycan",
    "glycordf",
]



import unittest

from glypy import monosaccharides, Substituent
from glypy.io.nomenclature import synonyms, identity


class IdentifyTests(unittest.TestCase):

    def test_is_a_predicate(self):
        for name, mono in monosaccharides.items():
            result = identity.is_a(mono, name)
            self.assertTrue(result)

    def test_identify_as(self):
        for name, mono in monosaccharides.items():
            if name in {"Hex", "Pen", "Oct", "Hep", "Non"}:
                continue
            pref_name = identity.identify(mono)
            if monosaccharides[pref_name] == mono:
                if not (name == pref_name or name in synonyms.monosaccharides[pref_name]):
                    raise AssertionError(
                        "{}".format((name, pref_name, synonyms.monosaccharides[pref_name])))

    def test_identify_index(self):
        identifier = identity.MonosaccharideIdentifier()
        for name, mono in monosaccharides.items():
            pref_name = identifier.identify(mono)
            if monosaccharides[pref_name] == mono:
                if not (name == pref_name or name in synonyms.monosaccharides[pref_name]):
                    raise AssertionError(
                        "{}".format((name, pref_name, synonyms.monosaccharides[pref_name])))

    def test_identify_substituents(self):
        self.assertTrue(
            identity.is_a(Substituent("n-acetyl"), Substituent("n-acetyl")))
        self.assertFalse(
            identity.is_a(Substituent('methyl'), Substituent('n-acetyl')))
        self.assertFalse(
            identity.is_a(monosaccharides.Man, Substituent('n-acetyl')))
        self.assertFalse(
            identity.is_a(Substituent('n-acetyl'), monosaccharides.Man))

    def test_get_preferred_name(self):
        self.assertTrue(identity.get_preferred_name('bdMan') == 'Man')

    def test_identify_failure(self):
        # Will fail because Hex is blacklisted
        self.assertRaises(
            identity.IdentifyException, lambda: identity.identify(monosaccharides.Hex))

    def test_precision(self):
        self.assertFalse(identity.is_a(monosaccharides.Kdn, monosaccharides.NeuAc))
        self.assertFalse(identity.is_a(monosaccharides.NeuAc, monosaccharides.Kdn))

    def test_grouping_axes(self):
        tree = identity.residue_list_to_tree((monosaccharides).values())


if __name__ == '__main__':
    unittest.main()



import sys
from setuptools import setup, find_packages, Extension

# With gratitude to the SqlAlchemy setup.py authors

from distutils.command.build_ext import build_ext
from distutils.errors import (CCompilerError, DistutilsExecError,
                              DistutilsPlatformError)

ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)
if sys.platform == 'win32':
    # 2.6's distutils.msvc9compiler can raise an IOError when failing to
    # find the compiler
    ext_errors += (IOError,)

try:
    from Cython.Build import cythonize
    extensions = cythonize(
        [Extension("glypy.composition.ccomposition", ["glypy/composition/ccomposition.pyx"]),
         Extension("glypy.utils.cenum", ['glypy/utils/cenum.pyx']),
         Extension("glypy._c.count_table", ["glypy/_c/count_table.pyx"]),
         ],
        annotate=True)
except (ImportError, AttributeError):
    print("No Cython")
    extensions = [
        Extension('glypy.composition.ccomposition', sources=[
                  'glypy/composition/ccomposition.c']),
        Extension("glypy.utils.cenum", ['glypy/utils/cenum.c']),
        Extension("glypy._c.count_table", ["glypy/_c/count_table.c"]),
    ]

cmdclass = {}


with open("glypy/version.py") as version_file:
    version = None
    for line in version_file.readlines():
        if "version = " in line:
            version = line.split(" = ")[1].replace("\"", "").strip()
            break
    else:
        print("Cannot determine version")

long_description = ''
try:
    with open("README.rst") as readme_file:
        long_description = readme_file.read()
except Exception:
    pass


class BuildFailed(Exception):

    def __init__(self):
        self.cause = sys.exc_info()[1]  # work around py 2/3 different syntax

    def __str__(self):
        return str(self.cause)


class ve_build_ext(build_ext):
    # This class allows C extension building to fail.

    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            raise BuildFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except ext_errors:
            raise BuildFailed()
        except ValueError:
            # this can happen on Windows 64 bit, see Python issue 7511
            if "'path'" in str(sys.exc_info()[1]):  # works with both py 2/3
                raise BuildFailed()
            raise


cmdclass['build_ext'] = ve_build_ext


def status_msgs(*msgs):
    print('*' * 75)
    for msg in msgs:
        print(msg)
    print('*' * 75)


required = ["hjson", "six"]


extras = {
    'plot': ["matplotlib>=2.2.0"],
    'glyspace': ['requests', 'rdflib', "SPARQLWrapper"]
}

extras['all'] = list({d for extra in extras.values() for d in extra})


def run_setup(include_cext=True):
    setup(
        name='glypy',
        version=version,
        packages=find_packages(),
        include_package_data=True,
        package_data={
            "glypy.structure": ["glypy/structure/data/*"],
            "glypy.io": ["glypy/io/data/*"],
            "glypy.io.nomenclature": ["glypy/io/nomenclature/data/*"],
        },
        install_requires=required,
        extras_require=extras,
        cmdclass=cmdclass,
        zip_safe=False,
        keywords="glycomics glycan carbohydrate glycoinformatics glypy n-linked o-linked glycosaminoglycan",
        description="A Glycoinformatics Toolkit",
        long_description=long_description,
        ext_modules=extensions if include_cext else None,
        url="https://github.com/mobiusklein/glypy",
        maintainer='Joshua Klein',
        maintainer_email="jaklein@bu.edu",
        project_urls={

        }
    )


try:
    run_setup(True)
except Exception as exc:
    status_msgs(
        str(exc),
        "WARNING: The C extension could not be compiled, " +
        "speedups are not enabled.",
        "Failure information, if any, is above.",
        "Retrying the build without the C extension now."
    )

    run_setup(False)

    status_msgs(
        "WARNING: The C extension could not be compiled, " +
        "speedups are not enabled.",
        "Plain-Python build succeeded."
    )



import pkg_resources
import json

try:
    from urllib import urlopen
except ImportError:
    from urllib.request import urlopen

from collections import defaultdict

from six import string_types as basestring

import glypy.io
from glypy.utils import StringIO
from glypy.utils.lazy import ProxyObject


class EnzymeCommissionNumber(object):
    """Represents an Enzyme Commission (E.C.) number specifying
    a unique identifier for a particular enzyme

    Attributes
    ----------
    activity : int
        The second-most precise number, describing the method of action
        from the particular :attr:`group`
    category : int
        The broadest number
    group : int
        The second broadest number
    identity : int
        The most precise number, describes which enzyme with the particular
        activity designated by :attr:`activity`
    """

    def __init__(self, category, group, activity, identity):
        self.category = int(category)
        self.group = int(group)
        self.activity = int(activity)
        self.identity = int(identity)
        self._str = '.'.join(map(str, self))

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash(self._str)

    def __ne__(self, other):
        return not (self == other)

    def __getitem__(self, i):
        if i == 0:
            return self.category
        elif i == 1:
            return self.group
        elif i == 2:
            return self.activity
        elif i == 3:
            return self.identity
        elif isinstance(i, slice) or i < 0:
            return tuple(self)[i]
        else:
            raise IndexError(i)

    def __iter__(self):
        yield self.category
        yield self.group
        yield self.activity
        yield self.identity

    def __str__(self):
        return self._str

    def __repr__(self):
        return ("{self.__class__.__name__}({self.category}, {self.group},"
                " {self.activity}, {self.identity})").format(self=self)

    @classmethod
    def parse(cls, string):
        string = str(string)
        parts = string.split(".")
        if len(parts) != 4:
            raise ValueError(
                "EC Numbers must have 4 parts (found %d): %r" % (len(parts, string)))
        try:
            parts = tuple(map(int, parts))
        except ValueError:
            raise ValueError("EC Numbers must be integers: %r" % (parts,))
        return cls(*parts)


class EnzymeInformation(object):
    """A collection of descriptive and identifying information about a single enzyme

    Attributes
    ----------
    alternative_names : :class:`list`
        Alternative names for the described enzyme
    ec_number : :class:`EnzymeCommisionNumber`
        The unique EC number for the described enzyme
    extra_information : :class:`dict`
        Arbitrary key-value information about described enzyme
    name : :class:`str`
        A human-readable name for the described enzyme
    """

    def __init__(self, name, ec_number=None, alternative_names=None, **kwargs):
        if isinstance(ec_number, basestring):
            ec_number = EnzymeCommissionNumber.parse(ec_number)
        elif isinstance(ec_number, (list, tuple)):
            ec_number = EnzymeCommissionNumber(*ec_number)

        self.name = name
        self.ec_number = ec_number
        self.alternative_names = alternative_names or ()
        self.extra_information = kwargs

    def __repr__(self):
        return ("{self.__class__.__name__}({self.name!r}, {self.ec_number!r}, "
                "{self.alternative_names!r}, **{self.extra_information!r})").format(self=self)

    def __eq__(self, other):
        if isinstance(other, basestring):
            return other in (self.name, self.ec_number) or other in self.alternative_names
        else:
            try:
                return other.ec_number == self.ec_number
            except AttributeError:
                return False

    def _to_dict(self):
        store = dict()
        store["name"] = self.name
        store['ec_number'] = str(self.ec_number)
        store['alternative_names'] = tuple(self.alternative_names)
        store['extra_information'] = dict(self.extra_information)
        return store

    def __hash__(self):
        return hash(self.name)


class EnzymeDatabase(object):
    _expasy_url = "ftp://ftp.expasy.org/databases/enzyme/enzyme.dat"

    def __init__(self, fp=None, format='json'):
        self.layered_store = defaultdict(
            lambda: defaultdict(lambda: defaultdict(dict)))
        self.direct_store = dict()

        if fp is not None:
            if format == 'expasy':
                self._parse_expasy(fp)
            elif format == 'json':
                self._parse_json(fp)
            else:
                raise ValueError("Unrecognized format: %r" % (format,))

    def add(self, enzyme_info):
        parts = enzyme_info.ec_number
        store = self.layered_store
        for k in parts[:-1]:
            store = store[k]
        store[parts[-1]] = enzyme_info
        self.direct_store[str(parts)] = enzyme_info

    def __getitem__(self, key):
        if isinstance(key[0], int):
            key = EnzymeCommissionNumber(*key)
        else:
            key = EnzymeCommissionNumber.parse(str(key))
        store = self.layered_store
        for k in key[:-1]:
            store = store[k]
        try:
            enzyme_info = store[key[-1]]
        except KeyError:
            enzyme_info = self.direct_store[str(key)]
        return enzyme_info

    def _parse_json(self, fp):
        enzymes = json.load(fp)

        for enz in enzymes:
            self.add(EnzymeInformation(**enz))

    def _dump(self, fp):
        json.dump([e._to_dict() for e in self.direct_store.values()],
                  fp, indent=2, sort_keys=True)

    @classmethod
    def _build(cls):
        fp = urlopen(cls._expasy_url)
        return cls(fp, "expasy")

    def _parse_expasy(self, fp):
        enzymes = []

        def new_store():
            return defaultdict(list)

        def postprocess(enzyme_info):
            enzyme_info['name'] = ' '.join(enzyme_info['name']).strip(".")
            enzyme_info['alternative_names'] = [
                n for n in ''.join(enzyme_info['alternative_names']).split(".") if n]
            up = enzyme_info['uniprot']
            up = [p.strip() for ent in up for p in ent.split(";") if p]
            enzyme_info['uniprot'] = up
            pro = enzyme_info['prosite']
            pro = [p.strip() for ent in pro for p in ent.split(";")
                   if p and p != 'PROSITE']
            enzyme_info['prosite'] = pro
            enzyme_info['catalytic_activity'] = ' '.join(
                enzyme_info['catalytic_activity'])
            comments = enzyme_info['comments']
            comments = ' '.join(
                map(str.strip, map(str.rstrip, comments))).split("-!-")
            enzyme_info['comments'] = ''.join(
                [c for c in comments if c.strip()])
            enzyme_info['ec_number'] = EnzymeCommissionNumber.parse(enzyme_info[
                                                                    'id'])
            return enzyme_info

        current_enzyme = new_store()
        for line in fp:
            line = line.strip()
            sigil = line[:2]
            line = line[5:]
            if sigil == '//':
                if current_enzyme['id']:
                    try:
                        enzymes.append(postprocess(current_enzyme))
                    except ValueError as e:
                        if "n" in current_enzyme['id']:
                            pass
                        else:
                            print(e, current_enzyme['name'])
                current_enzyme = new_store()
            elif sigil == 'ID':
                current_enzyme['id'] = line
            elif sigil == 'DE':
                current_enzyme['name'].append(line.strip(" "))
            elif sigil == 'AN':
                current_enzyme['alternative_names'].append(line.strip(" "))
            elif sigil == 'CA':
                current_enzyme['catalytic_activity'].append(line.strip(" "))
            elif sigil == 'CF':
                current_enzyme['cofactors'].append(line.strip(" "))
            elif sigil == 'CC':
                current_enzyme['comments'].append(line)
            elif sigil == 'PR':
                current_enzyme['prosite'].append(line)
            elif sigil == 'DR':
                current_enzyme['uniprot'].append(line)
            else:
                print(current_enzyme.get('id'), sigil, line)

        for enz in enzymes:
            self.add(EnzymeInformation(**enz))

    @classmethod
    def _from_static(cls):
        data_buffer = pkg_resources.resource_string(
            glypy.io.__name__, "data/enzyme.json")
        if isinstance(data_buffer, bytes):
            data_buffer = data_buffer.decode("utf-8")
        return cls(StringIO(data_buffer), format='json')


expasy_enzyme_db = ProxyObject(EnzymeDatabase._from_static)



from glypy.composition import Composition
from glypy.structure import constants

anomer_map = {
    'a': constants.Anomer.alpha,
    'b': constants.Anomer.beta,
    'o': constants.Anomer.uncyclized,
    'x': constants.Anomer.x
}


SuperClass = constants.SuperClass
superclass_map = {
    "sug".upper(): SuperClass["sug"],
    "tri".upper(): SuperClass["tri"],
    "tet".upper(): SuperClass["tet"],
    "pen".upper(): SuperClass["pen"],
    "hex".upper(): SuperClass["hex"],
    "hep".upper(): SuperClass["hep"],
    "oct".upper(): SuperClass["oct"],
    "non".upper(): SuperClass["non"],
    "dec".upper(): SuperClass["dec"],
    "s11".upper(): SuperClass["s11"],
    "s12".upper(): SuperClass["s12"],
    "s13".upper(): SuperClass["s13"],
    "s14".upper(): SuperClass["s14"],
    "s15".upper(): SuperClass["s15"],
    "s16".upper(): SuperClass["s16"],
    "s17".upper(): SuperClass["s17"],
    "s18".upper(): SuperClass["s18"],
    "s19".upper(): SuperClass["s19"],
    "s20".upper(): SuperClass["s20"],
    "x".upper(): SuperClass["x"],
}


link_replacement_composition_map = {
    "o": Composition(H=1),  # The source Carbon's Hydroxyl group loses its Hydrogen
    "d": Composition("OH"),  # The Hydroxyl group of the target Carbon is lost
    "h": Composition(H=1),  # The Hydrogen of the target Carbon is lost
    "n": Composition(H=1),  # Non-sugar unit recieves bond
    "x": Composition({})  # Unknown
}


linkage_type_map = {
    'o': constants.LinkageType.backbone_oxygen,
    'd': constants.LinkageType.backbone_oxygen,
    'h': constants.LinkageType.backbone_hydrogen,
    'n': constants.LinkageType.other,
    'x': constants.LinkageType.x
}


Modification = constants.Modification
modification_map = {
    'd': Modification.Deoxygenated,
    'a': Modification.Acidic,
    'aldi': Modification.Alditol,
    'keto': Modification.Ketone,
    'en': Modification.DoubleBond,
    'sp': Modification.SP,
    'sp2': Modification.SP2,
    'geminal': Modification.Geminal
}



import unittest
from .common import monosaccharides, link, glycoct, load
from glypy.composition.composition_transform import derivatize
from glypy.composition import Composition


class LinkTests(unittest.TestCase):

    def test_link_equality(self):
        parent = monosaccharides['Hex']
        child = monosaccharides['Hex']
        other = monosaccharides['Hex']
        link_1 = link.Link(parent, child, parent_position=3,
                           child_position=3, parent_loss='H', child_loss='OH')
        link_2 = link.Link(child, other, parent_position=6,
                           child_position=3, parent_loss='H', child_loss='OH')
        self.assertEqual(link_1, link_1)
        self.assertNotEqual(link_1, link_2)
        self.assertFalse(link_1 is None)

    def test_loss_composition(self):
        parent = monosaccharides['Hex']
        child = monosaccharides['Hex']

        link_1 = link.Link(parent, child, parent_position=3,
                           child_position=3, parent_loss='H', child_loss='OH')

        self.assertEqual(link_1.parent_loss, Composition(formula="H"))
        self.assertEqual(link_1.child_loss, Composition(formula="OH"))

    def test_break_and_reconnect(self):
        parent = monosaccharides['Hex']
        child = monosaccharides['Hex']

        link_1 = link.Link(parent, child, parent_position=3,
                           child_position=3, parent_loss='H', child_loss='OH')
        link_1.break_link(refund=True)
        self.assertTrue(len(parent.links[3]) == 0)
        self.assertTrue(len(child.links[3]) == 0)

        self.assertFalse(link_1.is_attached())
        self.assertFalse(link_1.is_attached(deep=True))

        link_1._reconnect(refund=True)
        self.assertTrue(len(parent.links[3]) == 1)
        self.assertTrue(len(child.links[3]) == 1)

        self.assertTrue(link_1.is_attached())
        self.assertTrue(link_1.is_attached(deep=True))

    def test_traversal(self):
        parent = monosaccharides['Hex']
        child = monosaccharides['Hex']

        link_1 = link.Link(parent, child, parent_position=3,
                           child_position=3, parent_loss='H', child_loss='OH')
        self.assertTrue(link_1.is_parent(parent))
        self.assertTrue(link_1.is_child(child))
        self.assertEqual(link_1.to(parent), child)
        self.assertEqual(link_1.to(child), parent)
        self.assertRaises(KeyError, lambda: link_1.to(1))


class AmbiguousLinkTests(unittest.TestCase):

    def test_link_site_no_overlap(self):

        ambig = '''
                RES
                1b:x-dglc-HEX-x:x
                2s:n-acetyl
                3b:b-dglc-HEX-1:5
                4s:n-acetyl
                5b:b-dman-HEX-1:5
                6b:a-dman-HEX-1:5
                7b:b-dglc-HEX-1:5
                8s:n-acetyl
                9b:b-dgal-HEX-1:5
                10b:a-dman-HEX-1:5
                11b:a-dman-HEX-1:5
                LIN
                1:1d(2+1)2n
                2:1o(4+1)3d
                3:3d(2+1)4n
                4:3o(4+1)5d
                5:5o(3|6+1)6d
                6:6o(2+1)7d
                7:7d(2+1)8n
                8:7o(4+1)9d
                9:5o(3|6+1)10d
                10:10o(3|6+1)11d
                '''
        exact = '''
                RES
                1b:x-dglc-HEX-1:5
                2s:n-acetyl
                3b:b-dglc-HEX-1:5
                4s:n-acetyl
                5b:b-dman-HEX-1:5
                6b:a-dman-HEX-1:5
                7b:b-dglc-HEX-1:5
                8s:n-acetyl
                9b:a-dman-HEX-1:5
                10b:a-dman-HEX-1:5
                11b:a-dman-HEX-1:5
                LIN
                1:1d(2+1)2n
                2:1o(4+1)3d
                3:3d(2+1)4n
                4:3o(4+1)5d
                5:5o(3+1)6d
                6:6o(2+1)7d
                7:7d(2+1)8n
                8:5o(6+1)9d
                9:9o(3+1)10d
                10:9o(6+1)11d
                '''
        ambig = glycoct.loads(ambig)
        exact = glycoct.loads(exact)
        ambig.set_reducing_end(True)
        exact.set_reducing_end(True)
        self.assertEqual(ambig.total_composition(), exact.total_composition())
        self.assertEqual(ambig, ambig.clone())
        self.assertEqual(derivatize(ambig.clone(), 'methyl').total_composition(),
                         derivatize(exact.clone(), 'methyl').total_composition())



# Much code here is derived from https://github.com/glycoinfo/glycocttowurcs
# though the representation of glycans in Eurocarbdb.MolecularFramework
# may not map 1:1.
try:
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence

import warnings

from six import string_types as basestring

from .basetype_conversion import (
    descriptors_to_base_type)

from glypy.structure.monosaccharide import Monosaccharide, ReducedEnd
from glypy.structure.constants import SuperClass, Anomer, Modification, Stem, Configuration, UnknownPosition
from glypy import OrderedMultiMap


anomer_map = {
    Anomer.beta: 'b',
    Anomer.alpha: 'a',
    Anomer.uncyclized: 'o',
    Anomer.x: 'x'
}


class CarbonDescriptors(Sequence):
    def __init__(self, descriptors, anomer, anomeric_position, ring_start, ring_end):
        self.descriptors = tuple(descriptors)
        self.anomer = Anomer[anomer]
        self.anomeric_position = self._translate_position(anomeric_position)
        self.ring_start = ring_start if ring_start is not None else UnknownPosition
        self.ring_end = ring_end if ring_end is not None else UnknownPosition

    def _translate_position(self, position):
        if position == '?':
            position = -1
        elif position == -1:
            position = '?'
        else:
            position = int(position)
        return position

    def __len__(self):
        return len(self.descriptors)

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, basestring):
            return str(self) == other
        if self.descriptors != other.descriptors:
            return False
        elif self.anomer != other.anomer:
            return False
        elif self.anomeric_position != other.anomeric_position:
            return False
        elif self.ring_start != other.ring_start:
            return False
        elif self.ring_end != other.ring_end:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.descriptors)

    def __getitem__(self, i):
        return self.descriptors[i]

    def __iter__(self):
        return iter(self.descriptors)

    def to_d_stereoform(self, code):
        out = []
        is_l_stereoform = code[-1] == '3'
        for site in code:
            if is_l_stereoform:
                if site == '3':
                    out.append('4')
                elif site == '4':
                    out.append('3')
                else:
                    out.append(site)
            else:
                out.append(site)
        return out

    def to_base_type(self):
        '''Convert the :class:`CarbonDescriptors` into a
        :class:`~.Monosaccharide`, not including substituents.

        Returns
        -------
        :class:`~.Monosaccharide`
        '''
        superclass = SuperClass[len(self)]
        carbon_coding = list(map(str, self))
        modifications = OrderedMultiMap()
        is_reduced = False
        # translate stereocode into generic carbon code
        for i, site in enumerate(carbon_coding):
            if site == '1':
                carbon_coding[i] = '3'
            elif site == '2':
                carbon_coding[i] = '4'
        start = 1
        stems = []
        configurations = []
        anomer = self.anomer
        ring_start = self.ring_start
        ring_end = self.ring_end
        if carbon_coding[0] == carbon_coding[-1] == 'h':
            anomer = Anomer.uncyclized
            ring_start = 0
            ring_end = 0
            is_reduced = True
        # if the stereosites are all defined
        if 'x' not in carbon_coding:
            # incrementally walk along the carbon sequence
            while start < superclass.value:
                # consider ring stereoforms of up to four carbons ahead, preferring longer
                # stereosequences,
                for i in range(4, 0, -1):
                    # extract the raw stereosequence
                    raw_chunk = carbon_coding[start:start + i]
                    # convert the stereosequence to D configuration and
                    # convert to a string for hash lookup
                    chunk = ''.join(self.to_d_stereoform(raw_chunk))
                    try:
                        # if the look up is successful
                        stem_name = descriptors_to_base_type[chunk]
                        # save the mapped stem name
                        stems.append(stem_name)
                        # infer the chirality of the ring from the last
                        # stereosite
                        conf = Configuration.x
                        if raw_chunk[-1] == '3':
                            conf = Configuration.l
                        elif raw_chunk[-1] == '4':
                            conf = Configuration.d
                        configurations.append(conf)
                        # start the lookup process again from the next starting
                        # location
                        start += len(raw_chunk)
                        break
                    except KeyError:
                        continue
                else:
                    # if no stereosequence could be detected, if the start position
                    # is a stereosite, then we may have a grolene trilose component
                    if chunk in ('3', '4'):
                        stems.append(descriptors_to_base_type['x'])
                        # infer the chirality of the ring from the last
                        # stereosite
                        conf = Configuration.x
                        if raw_chunk[-1] == '3':
                            conf = Configuration.l
                        elif raw_chunk[-1] == '4':
                            conf = Configuration.d
                        configurations.append(conf)
                    start += 1
        else:
            # This cannot handle unspecified nonulonic acids and other modified but unspecified
            # monosaccharides with multiple chiral centers well.
            stems.append(None)
            if carbon_coding[0] in ('u', 'h'):
                configurations.append(None)
            else:
                warnings.warn("Cannot infer chirality from %r" % (str(self),))
                configurations.append(None)
            # Guess if the monosaccharide is large enough to have a second chiral center, because
            # no other rule seems obvious. This could produce incorrect monosaccharide compositions?
            if len(carbon_coding) > 6:
                stems.append(None)
                configurations.append(None)

        anomeric_position = None
        double_bonds = []
        for i, site in enumerate(self):
            if site in ('a', 'u', 'U'):
                anomeric_position = i + 1
                if anomeric_position == 2:
                    modifications[anomeric_position] = Modification.keto
            if site in ('E', 'F'):
                double_bonds.append(i + 1)
            if site in ('d', 'm'):
                modifications[i + 1] = Modification.Deoxygenated
            if site == 'A':
                modifications[i + 1] = Modification.Acidic
        for site in double_bonds[::2]:
            modifications[i] = Modification.en
        stems = [Stem[x] for x in stems[::-1]]
        configurations = configurations[::-1]

        base = Monosaccharide(
            anomer,
            configurations,
            stems,
            superclass,
            ring_start,
            ring_end,
            modifications, reduced=ReducedEnd() if is_reduced else None)
        return base

    @classmethod
    def from_monosaccharide(cls, monosaccharide):
        '''Create a :class:`CarbonDescriptors` from a given
        :class:`~.Monosaccharide`.

        Parameters
        ----------
        monosaccharide: :class:`~.Monosaccharide`
            The monosaccharide to describe

        Returns
        -------
        :class:`CarbonDescriptors`
        '''
        code = ['x'] * monosaccharide.superclass.value
        stereocode = monosaccharide.stereocode
        code = [str(x.value) if x.value is not None else 'x' for x in stereocode]
        code[0] = 'u'
        code[-1] = 'h'
        if monosaccharide.anomer == 'uncyclized':
            code[0] = 'h'
            code[-1] = 'h'
        anomer = monosaccharide.anomer
        anomeric_position = monosaccharide.ring_start
        anomeric_sites = []
        is_aldose = True
        # encode the modifications onto the carbon descriptor code
        for position, modification in monosaccharide.modifications.items():
            is_terminal = (position == 1 or position == monosaccharide.superclass.value)
            if modification == Modification.Acidic:
                if not is_terminal:
                    raise ValueError("Cannot add a carboxylic acid group to a non-terminal carbon")
                if position == 1:
                    is_aldose = False
                code[position - 1] = 'A'
            elif modification == Modification.Deoxygenated:
                if position == 1:
                    is_aldose = False
                if is_terminal:
                    code[position - 1] = 'm'
                else:
                    code[position - 1] = 'd'
            elif modification == Modification.Ketone:
                is_aldose = False
                # code[position] = 'o'
                anomeric_sites.append(position)
            elif modification == Modification.en:
                code[position - 1] = 'E'
            elif modification == Modification.Alditol:
                is_aldose = False
                if position != 1:
                    raise ValueError("\"aldi\" must occur on the first carbon")
        if is_aldose:
            anomeric_sites.append(1)
        anomeric_position = anomeric_sites[0]
        # if the anomeric position is fully defined and the monosaccharide is cyclic
        if monosaccharide.ring_start not in (UnknownPosition, 0):
            code[anomeric_position - 1] = 'a'
        # if the anomeric position is partially undefined, the carbon code is 'u'
        elif monosaccharide.ring_start == UnknownPosition:
            code[anomeric_position - 1] = 'u'
        if monosaccharide.ring_start == UnknownPosition:
            anomeric_position = "?"
        return cls(code, anomer, anomeric_position, monosaccharide.ring_start, monosaccharide.ring_end)

    def to_backbone_code(self):
        '''Convert :class:`CarbonDescriptors` into a string representation
        matching the ``<BackboneCode>`` pattern from WURCS2.0

        Returns
        -------
        :class:`str`
        '''
        parts = []
        # carbon descriptors
        parts.append(''.join([i for i in self]))
        # if the anomer is completely undefined, do not include it
        if not (self.anomeric_position == -1 and self.anomer == Anomer.x):
            parts.append("-%s%s" % (self._translate_position(self.anomeric_position),
                                    anomer_map[self.anomer]))
        # if the ring is neither undefined nor open, include it
        if (self.ring_start != -1 and self.ring_end != -1) and (self.ring_start != 0 and self.ring_end != 0):
            parts.append("_%s-%s" % tuple(map(self._translate_position, (self.ring_start, self.ring_end))))
        return ''.join(parts)

    def __str__(self):
        return self.to_backbone_code()

    def __repr__(self):
        descriptors = ''.join(map(str, self))
        template = ("{self.__class__.__name__}({descriptors!r}, {self.anomer.name!r}, "
                    "{self.anomeric_position}, {self.ring_start}, {self.ring_end})")
        return template.format(self=self, descriptors=descriptors)



cimport cython
from cpython cimport PyErr_SetString, PyErr_SetObject
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem
from cpython.object cimport PyObject
from cpython.int cimport PyInt_AsLong


cdef str QMARK = "?"


cdef class EnumMeta(type):
    def __cinit__(self, name, parents, attrs):
        self.name_map = {}
        self.value_map = {}
        EnumType = attrs.get("__enum_type__", EnumValue)
        mapped = dict()
        for label, value in list(attrs.items()):
            if not (label.startswith("__") or label == "mro"):
                attrs.pop(label)
                delattr(self, label)
                enum_value = EnumType(self, label, value)
                if value in mapped:
                    mapped[value].add_name(label)
                    self.put(label, mapped[value])
                else:
                    mapped[value] = enum_value
                    self.put(label, enum_value)

        self.put(QMARK, None)
        for tp in parents:
            if isinstance(tp, EnumMeta):
                for label, value in tp:
                    if label == QMARK:
                        continue
                    self.put(label, value)

    def __dir__(self):
        return list(self.name_map.keys())

    def __iter__(self):
        for attr, val in self.name_map.items():
            if not attr.startswith("__") or attr == "mro":
                yield (attr, val)

    def __contains__(self, k):
        val = (k in self.name_map) or (k in self.value_map)
        return val

    def __getattr__(self, name):
        try:
            return self.translate(name)
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        self.put(name, value)

    def __getitem__(self, name):
        return self.translate(name)

    def __setitem__(self, name, value):
        self.put(name, value)

    cdef int put(self, name, value) except -1:
        if not isinstance(value, EnumValue):
            existing_value = self.get(value)
            if existing_value is not None:
                value = existing_value
                try:
                    value.add_name(name)
                except KeyError as e:
                    PyErr_SetObject(KeyError, str(e))
                    return -1
            else:
                value = EnumValue(self, name, value)
        PyDict_SetItem(self.name_map, name, value)
        PyDict_SetItem(self.value_map, value.value, name)
        return 0

    cdef EnumValue get(self, k):
        cdef:
            PyObject* presult
        presult = PyDict_GetItem(self.name_map, k)
        if presult == NULL:
            presult = PyDict_GetItem(self.value_map, k)
            if presult == NULL:
                return None
            else:
                presult = PyDict_GetItem(self.name_map, <object>presult)
                if presult == NULL:
                    return None
        return <EnumValue>presult

    cpdef translate(self, k):
        '''
        Attempt to translate the input object ``k`` into a data member of the Enum.

        First try to find an element of ``self`` by hashing it against member names.

        Then try to find an element of ``self`` by searching for a member in self that
        is value-equal to ``k``

        Otherwise throw a :exc:`KeyError`

        Parameters
        ----------
        k: object
            The value to be translated.

        Returns
        -------
        :class:`EnumValue`

        '''
        cdef:
            EnumValue result
        result = self.get(k)
        if result is None:
            raise KeyError("Could not translate {0} through {1}".format(k, self))
        return result

    cpdef EnumValue name(self, v):
        cdef:
            PyObject* presult
            object result
        presult = PyDict_GetItem(self.value_map, v)
        if presult == NULL:
            return None
        else:
            result = <object>presult
            return result

    def __repr__(self):
        return "<Enum {0}>".format(self.__name__)

    def __call__(self, v):
        return self.translate(v)


cdef class EnumValue(object):
    '''Represents a wrapper around an value with a name to identify it and
    more rich comparison logic. A value of an enumerated type'''

    def __init__(self, group, name, value, other_names=None):
        self.name = name
        self.value = value
        self.names = {name} | (other_names or set())
        self.group = group
        self._hash = None

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self.name)
        return self._hash

    def __index__(self):  # pragma: no cover
        return self.value

    def __int__(self):
        return int(self.value)

    def __eq__(self, other):
        cdef:
            EnumValue other_ev
            object temp
        if not isinstance(self, EnumValue):
            temp = self
            self = <EnumValue>other
            other = temp
        if isinstance(other, EnumValue):
            other_ev = <EnumValue>other
            if self.group is not other_ev.group:
                return False
            return self is other_ev
        else:
            return self.value == other or other in self.names

    def __and__(self, other):
        cdef:
            object temp
            EnumValue self_t
        if not isinstance(self, EnumValue):
            temp = self
            self_t = <EnumValue>other
            other = temp
        else:
            self_t = <EnumValue>self
        return self_t.value & other

    def __or__(self, other):
        cdef:
            object temp
            EnumValue self_t
        if not isinstance(self, EnumValue):
            temp = self
            self_t = <EnumValue>other
            other = temp
        else:
            self_t = <EnumValue>self
        return self_t.value | other

    def __xor__(self, other):
        cdef:
            object temp
            EnumValue self_t
        if not isinstance(self, EnumValue):
            temp = self
            self_t = <EnumValue>other
            other = temp
        else:
            self_t = <EnumValue>self
        return self_t.value ^ other

    def __ne__(self, other):
        return not self == other

    def __repr__(self):  # pragma: no cover
        return "<{group_name} {name}:{value}>".format(name=self.name,
                                                      group_name=self.group.__name__,
                                                      value=self.value)

    def __str__(self):
        return self.name

    def __reduce__(self):
        return self.group, (self.name,)

    cpdef add_name(self, basestring name, bint force=False):
        if name not in self.group or force:
            self.names.add(name)
            self.group[name] = self
        else:
            raise KeyError("{} already exists in {}".format(name, self.group))

    def resolve(self, mapping):
        for name in self.names:
            try:
                if name in mapping:
                    return mapping[name]
            except KeyError:  # pragma: no cover
                pass
        raise KeyError("Could not resolve {} against {}".format(self, mapping))

    def __lt__(self, other):
        return self.value < other.value

    def __gt__(self, other):
        return self.value > other.value

    def __le__(self, other):
        return self.value <= other.value

    def __ge__(self, other):
        return self.value >= other.value

    cpdef int int_value(self) except -1:
        try:
            return PyInt_AsLong(int(self.value))
        except Exception:
            PyErr_SetString(TypeError, "Could not convert %s to int" % (self.value, ))
            return -1


cdef class IntEnumValue(EnumValue):
    def __init__(self, group, name, value, other_names=None):
        super(IntEnumValue, self).__init__(group, name, value, other_names)
        self._int_value = PyInt_AsLong(self.value)

    cpdef int int_value(self) except -1:
        return self._int_value


import unittest

from glypy.io import glycoct
from glypy.structure.named_structures import glycans, monosaccharides
from glypy.algorithms import subtree_search

from .common import load


bokan_bao_gly_a = """RES
1b:b-dman-HEX-1:5
2b:a-dman-HEX-1:5
3b:a-dman-HEX-1:5
4b:b-dglc-HEX-1:5
5b:b-dglc-HEX-1:5
6s:n-acetyl
7b:b-dgal-HEX-1:5
8s:n-acetyl
9b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
10s:n-acetyl
LIN
1:1o(3+1)2d
2:1o(6+1)3d
3:2o(4+1)4d
4:3o(6+1)5d
5:4d(2+1)6n
6:4o(4+1)7d
7:5d(2+1)8n
8:7o(3+2)9d
9:9d(5+1)10n"""

bokan_bao_gly_b = """RES
1b:b-dman-HEX-1:5
2b:a-dman-HEX-1:5
3b:a-dman-HEX-1:5
4b:b-dglc-HEX-1:5
5b:b-dglc-HEX-1:5
6s:n-acetyl
7s:n-acetyl
8b:b-dgal-HEX-1:5
9b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
10s:n-acetyl
LIN
1:1o(3+1)2d
2:1o(6+1)3d
3:2o(4+1)4d
4:3o(6+1)5d
5:4d(2+1)6n
6:5d(2+1)7n
7:5o(4+1)8d
8:8o(3+2)9d
9:9d(5+1)10n"""


class SubtreeSearchTests(unittest.TestCase):

    def test_topological_inclusion(self):
        branchy = load("branchy_glycan")
        broad = load("broad_n_glycan")
        match = subtree_search.topological_inclusion(branchy.root, branchy.root)
        assert match == 10
        match = subtree_search.topological_inclusion(branchy.root, broad.get(3))
        assert match
        match = subtree_search.topological_inclusion(branchy.root, broad.root)
        assert not match

        gly_a = glycoct.loads(bokan_bao_gly_a)
        gly_b = glycoct.loads(bokan_bao_gly_b)
        assert subtree_search.topological_inclusion(gly_a.root, gly_b.root)
        assert subtree_search.topological_inclusion(gly_b.root, gly_a.root)

    def test_subtree_of(self):
        branchy = load("branchy_glycan")
        broad = load("broad_n_glycan")
        match = subtree_search.subtree_of(branchy, broad)
        assert match == 3
        core = glycans['N-Linked Core']
        self.assertTrue(subtree_search.subtree_of(core, broad))
        self.assertTrue(subtree_search.subtree_of(broad, core) is None)

    def test_maximum_common_subtree(self):
        core = glycans['N-Linked Core']
        branchy = load("branchy_glycan")
        res = subtree_search.maximum_common_subgraph(core, branchy)
        self.assertEqual(res.score, 6.0)

    def test_find_matching_subtree_roots(self):
        branchy = load("branchy_glycan")
        broad = load("broad_n_glycan")

        roots = subtree_search.find_matching_subtree_roots(branchy, broad)
        assert broad.get(3) in roots

    def test_is_n_glycan(self):
        core = glycans['N-Linked Core']
        tree = load("broad_n_glycan")
        result = (subtree_search.subtree_of(core, tree))
        self.assertTrue(result == 1)
        tree = load("complex_glycan")
        result = (subtree_search.subtree_of(core, tree, exact=False))
        self.assertTrue(result == 1)
        result = (subtree_search.subtree_of(core, tree, exact=True))
        self.assertTrue(result == 1)
        tree = load("branchy_glycan")
        result = (subtree_search.subtree_of(core, tree, exact=False))
        self.assertTrue(result is None)

    def test_disaccharide_similarity(self):
        core = glycans['N-Linked Core']
        self.assertEqual(subtree_search.n_saccharide_similarity(core, core), 1.0)
        copy = core.clone()
        copy.root.add_monosaccharide(monosaccharides.Fucose, 3)
        self.assertEqual(subtree_search.n_saccharide_similarity(core, copy), 0.7)

    def test_treelet_iterator(self):
        complex_glycan = load("complex_glycan")
        treelets = list(subtree_search.treelets(complex_glycan, 3))
        assert len(treelets) == 27



import zlib

try:
    from collections.abc import MutableSet
except ImportError:
    from collections import MutableSet

from glypy.io.glycoct import (loads, dumps)


class DistinctGlycanSet(MutableSet):
    """Store a distinct set of unique :class:`~.Glycan` objects
    space-efficiently.

    Implements the :class:`MutableSet` interface.

    This type stores structures by serializing the :class:`~.Glycan`
    into :title-reference:`GlycoCT{condensed}` text, and then compresses
    the text using :func:`zlib.compress`.

    Attributes
    ----------
    raw_data_buffer: :class:`set`
        The container for managing the unique glycans
    """

    def __init__(self, structures=None):
        if structures is None:
            structures = []
        self.raw_data_buffer = set()

        if isinstance(structures, DistinctGlycanSet):
            self.update(structures)
        else:
            for structure in structures:
                self.add(structure)

    def add(self, structure):
        """Add `structure` to the set

        Parameters
        ----------
        structure: :class:`~.Glycan`
            The structure to add to the set

        Returns
        -------
        :class:`bytes`
            The compressed encoding of `structure`
        """
        key = self._transform_text(
            self._structure_to_text(structure))
        if key in self.raw_data_buffer:
            return key
        self.raw_data_buffer.add(key)
        return key

    def discard(self, structure):
        """Remove `structure` from the set

        Parameters
        ----------
        structure: :class:`~.Glycan`
            The structure to remove
        """
        key = self._transform_text(
            self._structure_to_text(structure))
        self.raw_data_buffer.discard(key)

    def _structure_to_text(self, structure):
        return dumps(structure).encode('utf-8')

    def _text_to_structure(self, text):
        return loads(text.decode('utf-8'))

    def _transform_text(self, text):
        return zlib.compress(text)

    def _untransform_text(self, compressed):
        return zlib.decompress(compressed)

    def encode(self, structure):
        """Encode `structure` into compressed bytes

        Parameters
        ----------
        structure: :class:`~.Glycan`
            The structure to encode

        Returns
        -------
        :class:`bytes`
            The compressed bytes encoding `structure`
        """
        return self._transform_text(self._structure_to_text(structure))

    def add_encoded(self, encoded):
        """Add a pre-encoded structure to the set

        Parameters
        ----------
        encoded: :class:`bytes`
            An encoded structure
        """
        self.raw_data_buffer.add(encoded)

    def discard_encoded(self, encoded):
        """Remove a pre-encoded structure from the set

        Parameters
        ----------
        encoded: :class:`bytes`
            An encoded structure
        """
        self.raw_data_buffer.discard(encoded)

    def has_encoded(self, encoded):
        return encoded in self.raw_data_buffer

    def pop(self):
        text = self._untransform_text(self.raw_data_buffer.pop())
        return self._text_to_structure(text)

    def __len__(self):
        return len(self.raw_data_buffer)

    def __iter__(self):
        for compressed in self.raw_data_buffer:
            yield self._text_to_structure(
                self._untransform_text(compressed))

    def __contains__(self, structure):
        text = self._structure_to_text(structure)
        compressed = self._transform_text(text)
        return compressed in self.raw_data_buffer

    @classmethod
    def from_buffer_slice(cls, buffer_slice):
        inst = cls()
        inst.raw_data_buffer.update(buffer_slice)
        return inst

    def update(self, other):
        if isinstance(other, DistinctGlycanSet):
            self.raw_data_buffer.update(other.raw_data_buffer)
        else:
            for x in other:
                self.add(x)

    def partition(self, nchunks=2):
        buffer_sequence = list(self.raw_data_buffer)
        n = len(buffer_sequence)
        chunk_size = max(int(n / nchunks), 1)
        i = 0
        chunks = []
        while i < n:
            chunk = buffer_sequence[i:i + chunk_size]
            i += chunk_size
            chunks.append(self.from_buffer_slice(chunk))
        return chunks

    def remove_all(self, other):
        if not isinstance(other, DistinctGlycanSet):
            other = DistinctGlycanSet(other)
        self.raw_data_buffer -= other.raw_data_buffer

    def __sub__(self, other):
        return self.from_buffer_slice(
            self.raw_data_buffer - other.raw_data_buffer)

    def __isub__(self, other):
        self.raw_data_buffer -= other.raw_data_buffer
        return self

    def __and__(self, other):
        return self.from_buffer_slice(
            self.raw_data_buffer & other.raw_data_buffer)

    def __iand__(self, other):
        self.raw_data_buffer &= other.raw_data_buffer
        return self

    def __or__(self, other):
        return self.from_buffer_slice(
            self.raw_data_buffer | other.raw_data_buffer)

    def __ior__(self, other):
        self.raw_data_buffer.update(other.raw_data_buffer)
        return self



'''Implements a reader and writer for the `WURCS 2.0` format.
'''

from glypy.io.file_utils import ParserInterface
from .parser import loads
from .node_type import NodeTypeSpec
from .writer import dumps
from .utils import WURCSError
from .carbon_descriptors import CarbonDescriptors


class WURCSParser(ParserInterface):
    def process_result(self, line):
        structure = loads(line)
        return structure


__all__ = [
    "WURCSParser", "loads", "dumps",
    "NodeTypeSpec", "CarbonDescriptors",
    "WURCSError",
]



import unittest

from glypy.io import linear_code
from glypy.tests.common import load


class LinearCodeTests(unittest.TestCase):

    def test_translate(self):
        broad = load("broad_n_glycan")
        dup = linear_code.loads(linear_code.dumps(broad))
        self.assertEqual(broad, dup)

        # linear code doesn't know about modifications or
        # ring shape
        sulfated = load("sulfated_glycan")
        sulfated.reducing_end = None
        sulfated.root.ring_start = 1
        sulfated.root.ring_end = 5
        dup2 = linear_code.loads(linear_code.dumps(sulfated))
        self.assertEqual(dup2, sulfated)

        sulfated = load("sulfated_glycan")
        dup = linear_code.loads(linear_code.dumps(sulfated))
        self.assertNotEqual(sulfated, dup)

    def test_example(self):
        seq = 'NNa3Ab3(NNa6)AN'
        structure = linear_code.loads(seq)
        self.assertEqual(len(structure), 4)


if __name__ == '__main__':
    unittest.main()



import unittest

import glypy
from glypy.io import iupac
from glypy.enzyme import (
    MultiprocessingGlycome, make_n_glycan_pathway, EnzymeGraph)


class GlycomeTests(unittest.TestCase):

    def _make_glycome(self):
        glycosylases, glycosyltransferases, seeds = make_n_glycan_pathway()
        glycosyltransferases.pop("gntE")
        glycosyltransferases.pop('agal13galt')
        glycosyltransferases.pop('siat2_3')
        glycosyltransferases.pop('siat2_6')
        glycosyltransferases.pop("fuct3")
        glycome = MultiprocessingGlycome(
            glycosylases, glycosyltransferases, seeds)
        return glycome

    def test_path_between(self):
        with open("test_data/enzyme_graph.json", 'rt') as fh:
            graph = EnzymeGraph.load(fh)
        p = list(graph.seeds)[0]
        c = sorted(graph)[-1].child
        path = graph.path_between(p, c)
        assert len(path) == 13

    def test_synthesize_glycome(self):
        glycome = self._make_glycome()
        for i, gen in enumerate(glycome.run()):
            pass
        assert i == 20
        assert len(gen) == 1
        eg = EnzymeGraph(glycome.enzyme_graph)
        assert eg.node_count() == 498
        assert eg.edge_count() == 1427
        assert eg.parentless() == eg.seeds

        with open("test_data/enzyme_graph.json", 'rt') as fh:
            graph = EnzymeGraph.load(fh)
        assert eg == graph
        assert eg._dump() == graph._dump()
        seed = list(eg.seeds)[0]
        ref = list(graph.seeds)[0]
        assert seed == ref
        assert set(eg.children(seed)) == set(graph.children(seed))

    def test_galt(self):
        _, glycosyltransferases, _ = make_n_glycan_pathway()
        galt = glycosyltransferases['galt']
        target = iupac.loads(
            "a-D-Manp-(1-6)-[b-D-Glcp2NAc-(1-2)-a-D-Manp-(1-3)]b-D-Manp-(1-4)-b-D-Glcp2NAc-(1-4)-b-D-Glcp2NAc")
        for instance in galt(target):
            res = instance.mass() - glypy.monosaccharide_residues.Gal.mass()
            self.assertAlmostEqual(res, target.mass())


if __name__ == '__main__':
    unittest.main()



from .base import (opener, make_counter, invert_dict, identity,
                   nullop, chrinc, make_struct, classproperty, cyclewarning,
                   root, tree, groupby, pickle, ET, StringIO, where, uid,
                   basestring, RootProtocolNotSupportedError, TreeProtocolNotSupportedError)

from .enum import Enum

__all__ = ['opener', 'make_counter', 'invert_dict', 'identity', 'nullop',
           "chrinc", "make_struct", "classproperty", "cyclewarning",
           "root", "tree", "groupby", "uid", "Enum", "RootProtocolNotSupportedError",
           "TreeProtocolNotSupportedError"]
